# stdlib


def json_diff(json1: dict, json2: dict) -> dict:
    """
    Calculate the difference between two JSON objects and return the differences as a JSON object.

    Args:
        json1 (dict): The first JSON object.
        json2 (dict): The second JSON object.

    Returns:
        dict: A JSON object representing the differences.
    """

    def compare_dicts(d1: dict, d2: dict) -> dict:
        diffs = {}

        # Keys in d1 but not in d2 (deleted)
        for key in d1.keys() - d2.keys():
            diffs[key] = {"status": "deleted", "value": d1[key]}

        # Keys in d2 but not in d1 (added)
        for key in d2.keys() - d1.keys():
            diffs[key] = {"status": "added", "value": d2[key]}

        # Keys in both, but with different values (updated)
        for key in d1.keys() & d2.keys():
            if isinstance(d1[key], dict) and isinstance(d2[key], dict):
                # Recursively compare nested dictionaries
                nested_diff_dict = compare_dicts(d1[key], d2[key])
                if nested_diff_dict:
                    diffs[key] = nested_diff_dict
            elif isinstance(d1[key], list) and isinstance(d2[key], list):
                # Compare lists
                nested_diff_list = compare_lists(d1[key], d2[key])
                if nested_diff_list:
                    diffs[key] = nested_diff_list  # type: ignore
            elif d1[key] != d2[key]:
                diffs[key] = {
                    "status": "updated",
                    "old_value": d1[key],
                    "new_value": d2[key],
                }

        return diffs

    def compare_lists(l1: list, l2: list) -> list | None:
        diffs = []
        max_len = max(len(l1), len(l2))

        for i in range(max_len):
            if i >= len(l1):
                diffs.append({"status": "added", "value": l2[i]})
            elif i >= len(l2):
                diffs.append({"status": "deleted", "value": l1[i]})
            elif isinstance(l1[i], dict) and isinstance(l2[i], dict):
                # Recursively compare dictionaries in lists
                nested_diff_dict = compare_dicts(l1[i], l2[i])
                if nested_diff_dict:
                    diffs.append(nested_diff_dict)
            elif isinstance(l1[i], list) and isinstance(l2[i], list):
                # Recursively compare nested lists
                nested_diff_list = compare_lists(l1[i], l2[i])
                if nested_diff_list:
                    diffs.append(nested_diff_list)
            elif l1[i] != l2[i]:
                diffs.append(
                    {"status": "updated", "old_value": l1[i], "new_value": l2[i]}
                )

        return diffs if diffs else None

    # Generate the JSON diff
    json_diff_result = compare_dicts(json1, json2)

    return json_diff_result



# stdlib

# stdlib
import uuid

# third party
import sqlalchemy as sa
from sqlalchemy import Column
from sqlalchemy import Dialect
from sqlalchemy import Table
from sqlalchemy import TypeDecorator
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.types import JSON

# relative
from ...types.syft_object import SyftObject
from ...types.uid import UID


class SQLiteBase(DeclarativeBase):
    pass


class PostgresBase(DeclarativeBase):
    pass


class UIDTypeDecorator(TypeDecorator):
    """Converts between Syft UID and UUID."""

    impl = sa.UUID
    cache_ok = True

    def process_bind_param(self, value, dialect):  # type: ignore
        if value is not None:
            return value.value

    def process_result_value(self, value, dialect):  # type: ignore
        if value is not None:
            return UID(value)


def create_table(
    object_type: type[SyftObject],
    dialect: Dialect,
) -> Table:
    """Create a table for a given SYftObject type, and add it to the metadata.

    To create the table on the database, you must call `Base.metadata.create_all(engine)`.

    Args:
        object_type (type[SyftObject]): The type of the object to create a table for.
        dialect (Dialect): The dialect of the database.

    Returns:
        Table: The created table.
    """
    table_name = object_type.__canonical_name__
    dialect_name = dialect.name

    fields_type = JSON if dialect_name == "sqlite" else postgresql.JSON
    permissions_type = JSON if dialect_name == "sqlite" else postgresql.JSONB
    storage_permissions_type = JSON if dialect_name == "sqlite" else postgresql.JSONB

    Base = SQLiteBase if dialect_name == "sqlite" else PostgresBase

    if table_name not in Base.metadata.tables:
        Table(
            object_type.__canonical_name__,
            Base.metadata,
            Column("id", UIDTypeDecorator, primary_key=True, default=uuid.uuid4),
            Column("fields", fields_type, default={}),
            Column("permissions", permissions_type, default=[]),
            Column(
                "storage_permissions",
                storage_permissions_type,
                default=[],
            ),
            Column(
                "_created_at", sa.DateTime, server_default=sa.func.now(), index=True
            ),
            Column("_updated_at", sa.DateTime, server_onupdate=sa.func.now()),
            Column("_deleted_at", sa.DateTime, index=True),
        )

    return Base.metadata.tables[table_name]



# stdlib
from collections.abc import Callable
from enum import Enum
import logging
import secrets
from typing import Any
from typing import cast

# relative
from ...abstract_server import ServerType
from ...client.client import HTTPConnection
from ...client.client import PythonConnection
from ...client.client import SyftClient
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...server.worker_settings import WorkerSettings
from ...service.settings.settings import ServerSettings
from ...store.db.db import DBManager
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.server_url import ServerURL
from ...types.transforms import TransformContext
from ...types.transforms import keep
from ...types.transforms import make_set_default
from ...types.transforms import transform
from ...types.transforms import transform_method
from ...types.uid import UID
from ...util.util import generate_token
from ...util.util import get_env
from ...util.util import prompt_warning_message
from ...util.util import str_to_bool
from ..context import AuthedServiceContext
from ..metadata.server_metadata import ServerMetadata
from ..request.request import Request
from ..request.request import RequestStatus
from ..request.request import SubmitRequest
from ..response import SyftInfo
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..warnings import CRUDWarning
from .association_request import AssociationRequestChange
from .reverse_tunnel_service import ReverseTunnelService
from .routes import HTTPServerRoute
from .routes import PythonServerRoute
from .routes import ServerRoute
from .routes import ServerRouteType
from .server_peer import ServerPeer
from .server_peer import ServerPeerUpdate

logger = logging.getLogger(__name__)

REVERSE_TUNNEL_ENABLED = "REVERSE_TUNNEL_ENABLED"


def reverse_tunnel_enabled() -> bool:
    return str_to_bool(get_env(REVERSE_TUNNEL_ENABLED, "false"))


@serializable(canonical_name="ServerPeerAssociationStatus", version=1)
class ServerPeerAssociationStatus(Enum):
    PEER_ASSOCIATED = "PEER_ASSOCIATED"
    PEER_ASSOCIATION_PENDING = "PEER_ASSOCIATION_PENDING"
    PEER_NOT_FOUND = "PEER_NOT_FOUND"


@serializable(canonical_name="NetworkSQLStash", version=1)
class NetworkStash(ObjectStash[ServerPeer]):
    @as_result(StashException, NotFoundException)
    def get_by_name(self, credentials: SyftVerifyKey, name: str) -> ServerPeer:
        try:
            return self.get_one(
                credentials=credentials,
                filters={"name": name},
            ).unwrap()
        except NotFoundException as e:
            raise NotFoundException.from_exception(
                e, public_message=f"ServerPeer with {name} not found"
            )

    @as_result(StashException)
    def create_or_update_peer(
        self, credentials: SyftVerifyKey, peer: ServerPeer
    ) -> ServerPeer:
        """
        Update the selected peer and its route priorities if the peer already exists
        If the peer does not exist, simply adds it to the database.

        Args:
            credentials (SyftVerifyKey): The credentials used to authenticate the request.
            peer (ServerPeer): The peer to be updated or added.

        Returns:
            ServerPeer: The updated or added peer if the operation was successful.
                Raises an exception if the operation failed.
        """
        self.check_type(peer, ServerPeer).unwrap()

        try:
            existing_peer: ServerPeer = self.get_by_uid(
                credentials=credentials, uid=peer.id
            ).unwrap()
        except SyftException:
            return self.set(credentials, obj=peer).unwrap()

        existing_peer.update_routes(peer.server_routes)
        peer_update = ServerPeerUpdate(
            id=peer.id, server_routes=existing_peer.server_routes
        )
        return self.update(credentials, peer_update).unwrap()

    @as_result(StashException, NotFoundException)
    def get_by_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey
    ) -> ServerPeer:
        return self.get_one(
            credentials=credentials,
            filters={"verify_key": verify_key},
        ).unwrap()

    @as_result(StashException)
    def get_by_server_type(
        self, credentials: SyftVerifyKey, server_type: ServerType
    ) -> list[ServerPeer]:
        return self.get_all(
            credentials=credentials,
            filters={"server_type": server_type},
        ).unwrap()


@serializable(canonical_name="NetworkService", version=1)
class NetworkService(AbstractService):
    stash: NetworkStash

    def __init__(self, store: DBManager) -> None:
        self.stash = NetworkStash(store=store)
        if reverse_tunnel_enabled():
            self.rtunnel_service = ReverseTunnelService()

    @service_method(
        path="network.exchange_credentials_with",
        name="exchange_credentials_with",
        roles=GUEST_ROLE_LEVEL,
        warning=CRUDWarning(confirmation=True),
        unwrap_on_success=False,
    )
    def exchange_credentials_with(
        self,
        context: AuthedServiceContext,
        self_server_route: ServerRoute,
        remote_server_route: ServerRoute,
        remote_server_verify_key: SyftVerifyKey,
        reverse_tunnel: bool = False,
    ) -> Request | SyftSuccess:
        """
        Exchange Route With Another Server. If there is a pending association request, return it
        """

        # Step 1: Validate the Route
        self_server_peer = self_server_route.validate_with_context(
            context=context
        ).unwrap()

        if reverse_tunnel and not reverse_tunnel_enabled():
            raise SyftException(
                public_message="Reverse tunneling is not enabled on this server."
            )

        elif reverse_tunnel:
            _rtunnel_route = self_server_peer.server_routes[-1]
            _rtunnel_route.rtunnel_token = generate_token()
            _rtunnel_route.host_or_ip = f"{self_server_peer.name}.syft.local"
            self_server_peer.server_routes[-1] = _rtunnel_route

        # Step 2: Send the Server Peer to the remote server
        # Also give them their own to validate that it belongs to them
        # random challenge prevents replay attacks
        remote_client: SyftClient = remote_server_route.client_with_context(
            context=context
        )
        remote_server_peer = ServerPeer.from_client(remote_client)

        # Step 3: Check remotely if the self server already exists as a peer
        # Update the peer if it exists, otherwise add it
        try:
            remote_self_server_peer = (
                remote_client.api.services.network.get_peer_by_name(
                    name=self_server_peer.name
                )
            )
        except SyftException:
            remote_self_server_peer = None

        association_request_approved = True
        if isinstance(remote_self_server_peer, ServerPeer):
            updated_peer = ServerPeerUpdate(
                id=self_server_peer.id, server_routes=self_server_peer.server_routes
            )
            try:
                remote_client.api.services.network.update_peer(peer_update=updated_peer)
            except Exception as e:
                logger.error(f"Failed to update peer information on remote client. {e}")
                raise SyftException.from_exception(
                    e,
                    public_message=f"Failed to add peer information on remote client : {remote_client.id}",
                    private_message=f"Error: {e}",
                )

        # If peer does not exist, ask the remote client to add this server
        # (represented by `self_server_peer`) as a peer
        if remote_self_server_peer is None:
            random_challenge = secrets.token_bytes(16)
            remote_res = remote_client.api.services.network.add_peer(
                peer=self_server_peer,
                challenge=random_challenge,
                self_server_route=remote_server_route,
                verify_key=remote_server_verify_key,
            )
            association_request_approved = not isinstance(remote_res, Request)

        # Step 4: Save the remote peer for later
        self.stash.create_or_update_peer(
            context.server.verify_key,
            remote_server_peer,
        ).unwrap(public_message=f"Failed to save peer: {remote_server_peer}.")
        # Step 5: Save config to enable reverse tunneling
        if reverse_tunnel and reverse_tunnel_enabled():
            self.set_reverse_tunnel_config(
                context=context,
                self_server_peer=self_server_peer,
                remote_server_peer=remote_server_peer,
            )

        return (
            SyftSuccess(message="Routes Exchanged")
            if association_request_approved
            else remote_res
        )

    @service_method(path="network.add_peer", name="add_peer", roles=GUEST_ROLE_LEVEL)
    def add_peer(
        self,
        context: AuthedServiceContext,
        peer: ServerPeer,
        challenge: bytes,
        self_server_route: ServerRoute,
        verify_key: SyftVerifyKey,
    ) -> Request | SyftSuccess:
        """Add a Network Server Peer. Called by a remote server to add
        itself as a peer for the current server.
        """
        # Using the verify_key of the peer to verify the signature
        # It is also our single source of truth for the peer
        if peer.verify_key != context.credentials:
            raise SyftException(
                public_message=(
                    f"The {type(peer).__name__}.verify_key: "
                    f"{peer.verify_key} does not match the signature of the message"
                )
            )

        if verify_key != context.server.verify_key:
            raise SyftException(
                public_message="verify_key does not match the remote server's verify_key for add_peer"
            )

        # check if the peer already is a server peer
        existing_peer_res = self.stash.get_by_uid(context.server.verify_key, peer.id)

        if existing_peer_res.is_ok() and isinstance(
            existing_peer := existing_peer_res.ok(), ServerPeer
        ):
            msg = [
                f"The peer '{peer.name}' is already associated with '{context.server.name}'"
            ]

            if existing_peer != peer:
                msg.append("Peer information change detected.")
                self.stash.create_or_update_peer(
                    context.server.verify_key,
                    peer,
                ).unwrap(
                    public_message="\n".join(
                        msg + ["Attempt to update peer information failed."]
                    )
                )

                msg.append("Peer information successfully updated.")
                return SyftSuccess(message="\n".join(msg))

            return SyftSuccess(message="\n".join(msg))

        # check if the peer already submitted an association request
        association_requests: list[Request] = self._get_association_requests_by_peer_id(
            context=context, peer_id=peer.id
        )
        if (
            association_requests
            and (association_request := association_requests[-1]).status
            == RequestStatus.PENDING
        ):
            return association_request
        # only create and submit a new request if there is no requests yet
        # or all previous requests have been rejected
        association_request_change = AssociationRequestChange(
            self_server_route=self_server_route, challenge=challenge, remote_peer=peer
        )
        submit_request = SubmitRequest(
            changes=[association_request_change],
            requesting_user_verify_key=context.credentials,
        )
        request = context.server.services.request.submit(context, submit_request)
        if (
            isinstance(request, Request)
            and context.server.settings.association_request_auto_approval
        ):
            return context.server.services.request.apply(context, uid=request.id)

        return request

    @service_method(path="network.ping", name="ping", roles=GUEST_ROLE_LEVEL)
    def ping(self, context: AuthedServiceContext, challenge: bytes) -> bytes:
        """To check alivesness/authenticity of a peer"""

        # # Only the root user can ping the server to check its state
        # if context.server.verify_key != context.credentials:
        #     return SyftError(message=("Only the root user can access ping endpoint"))

        # this way they can match up who we are with who they think we are
        # Sending a signed messages for the peer to verify

        challenge_signature = context.server.signing_key.signing_key.sign(
            challenge
        ).signature

        return challenge_signature

    @service_method(
        path="network.check_peer_association",
        name="check_peer_association",
        roles=GUEST_ROLE_LEVEL,
    )
    def check_peer_association(
        self, context: AuthedServiceContext, peer_id: UID
    ) -> ServerPeerAssociationStatus:
        """Check if a peer exists in the network stash"""

        # get the server peer for the given sender peer_id
        try:
            self.stash.get_by_uid(context.server.verify_key, peer_id).unwrap()
            return ServerPeerAssociationStatus.PEER_ASSOCIATED
        except SyftException:
            association_requests: list[Request] = (
                self._get_association_requests_by_peer_id(
                    context=context, peer_id=peer_id
                )
            )
            if (
                association_requests
                and association_requests[-1].status == RequestStatus.PENDING
            ):
                return ServerPeerAssociationStatus.PEER_ASSOCIATION_PENDING

        return ServerPeerAssociationStatus.PEER_NOT_FOUND

    @service_method(
        path="network.get_all_peers", name="get_all_peers", roles=GUEST_ROLE_LEVEL
    )
    def get_all_peers(self, context: AuthedServiceContext) -> list[ServerPeer]:
        """Get all Peers"""
        return self.stash.get_all(
            credentials=context.server.verify_key,
            order_by="name",
            sort_order="asc",
        ).unwrap()

    @service_method(
        path="network.get_peer_by_name", name="get_peer_by_name", roles=GUEST_ROLE_LEVEL
    )
    def get_peer_by_name(self, context: AuthedServiceContext, name: str) -> ServerPeer:
        """Get Peer by Name"""
        return self.stash.get_by_name(
            credentials=context.server.verify_key,
            name=name,
        ).unwrap()

    @service_method(
        path="network.get_peers_by_type",
        name="get_peers_by_type",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_peers_by_type(
        self, context: AuthedServiceContext, server_type: ServerType
    ) -> list[ServerPeer]:
        return self.stash.get_by_server_type(
            credentials=context.server.verify_key,
            server_type=server_type,
        ).unwrap()

    @service_method(
        path="network.update_peer",
        name="update_peer",
        roles=GUEST_ROLE_LEVEL,
    )
    def update_peer(
        self,
        context: AuthedServiceContext,
        peer_update: ServerPeerUpdate,
    ) -> SyftSuccess:
        # try setting all fields of ServerPeerUpdate according to ServerPeer

        peer = self.stash.update(
            credentials=context.server.verify_key,
            obj=peer_update,
        ).unwrap()

        self.set_reverse_tunnel_config(context=context, remote_server_peer=peer)
        return SyftSuccess(
            message=f"Peer '{peer.name}' information successfully updated."
        )

    def set_reverse_tunnel_config(
        self,
        context: AuthedServiceContext,
        remote_server_peer: ServerPeer,
        self_server_peer: ServerPeer | None = None,
    ) -> None:
        server_type = cast(ServerType, context.server.server_type)
        if server_type.value == ServerType.GATEWAY.value:
            rtunnel_route = remote_server_peer.get_rtunnel_route()
            (
                self.rtunnel_service.set_server_config(remote_server_peer)
                if rtunnel_route
                else None
            )
        else:
            self_server_peer = (
                context.server.settings.to(ServerPeer)
                if self_server_peer is None
                else self_server_peer
            )
            rtunnel_route = self_server_peer.get_rtunnel_route()
            (
                self.rtunnel_service.set_client_config(
                    self_server_peer=self_server_peer,
                    remote_server_route=remote_server_peer.pick_highest_priority_route(),
                )
                if rtunnel_route
                else None
            )

    @service_method(
        path="network.delete_peer_by_id",
        name="delete_peer_by_id",
        roles=DATA_OWNER_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def delete_peer_by_id(self, context: AuthedServiceContext, uid: UID) -> SyftSuccess:
        """Delete Server Peer"""
        peer_to_delete = self.stash.get_by_uid(context.credentials, uid).unwrap()

        server_side_type = cast(ServerType, context.server.server_type)
        if server_side_type.value == ServerType.GATEWAY.value:
            rtunnel_route = peer_to_delete.get_rtunnel_route()
            (
                self.rtunnel_service.clear_server_config(peer_to_delete)
                if rtunnel_route
                else None
            )

        # TODO: Handle the case when peer is deleted from datasite server

        self.stash.delete_by_uid(context.credentials, uid).unwrap()

        # Delete all the association requests from this peer
        association_requests: list[Request] = self._get_association_requests_by_peer_id(
            context=context, peer_id=uid
        )
        for request in association_requests:
            context.server.services.request.delete_by_uid(context, request.id)
        # TODO: Notify the peer (either by email or by other form of notifications)
        # that it has been deleted from the network
        return SyftSuccess(message=f"Server Peer with id {uid} deleted.")

    @service_method(
        path="network.add_route_on_peer",
        name="add_route_on_peer",
        unwrap_on_success=False,
    )
    def add_route_on_peer(
        self,
        context: AuthedServiceContext,
        peer: ServerPeer,
        route: ServerRoute,
    ) -> SyftSuccess:
        """
        Add or update the route information on the remote peer.

        Args:
            context (AuthedServiceContext): The authentication context.
            peer (ServerPeer): The peer representing the remote server.
            route (ServerRoute): The route to be added.

        Returns:
            SyftSuccess: A success message if the route is verified,
                otherwise an error message.
        """
        # creates a client on the remote server based on the credentials
        # of the current server's client
        remote_client = peer.client_with_context(context=context).unwrap()
        # ask the remote server to add the route to the self server
        result = remote_client.api.services.network.add_route(
            peer_verify_key=context.credentials,
            route=route,
            called_by_peer=True,
        )

        return SyftSuccess(
            message="Route information added to remote peer", value=result
        )

    @service_method(
        path="network.add_route",
        name="add_route",
        roles=GUEST_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def add_route(
        self,
        context: AuthedServiceContext,
        peer_verify_key: SyftVerifyKey,
        route: ServerRoute,
        called_by_peer: bool = False,
    ) -> SyftSuccess:
        """
        Add a route to the peer. If the route already exists, update its priority.

        Args:
            context (AuthedServiceContext): The authentication context of the remote server.
            peer_verify_key (SyftVerifyKey): The verify key of the remote server peer.
            route (ServerRoute): The route to be added.
            called_by_peer (bool): The flag to indicate that it's called by a remote peer.

        Returns:
            SyftSuccess
        """
        # verify if the peer is truly the one sending the request to add the route to itself
        if called_by_peer and peer_verify_key != context.credentials:
            raise SyftException(
                public_message=(
                    f"The {type(peer_verify_key).__name__}: "
                    f"{peer_verify_key} does not match the signature of the message"
                )
            )
        # get the full peer object from the store to update its routes
        remote_server_peer: ServerPeer = (
            self._get_remote_server_peer_by_verify_key(context, peer_verify_key)
        ).unwrap()
        # add and update the priority for the peer
        if route in remote_server_peer.server_routes:
            return SyftSuccess(
                message=f"The route already exists between '{context.server.name}' and "
                f"peer '{remote_server_peer.name}'."
            )

        remote_server_peer.update_route(route=route)

        # update the peer in the store with the updated routes
        peer_update = ServerPeerUpdate(
            id=remote_server_peer.id, server_routes=remote_server_peer.server_routes
        )
        self.stash.update(
            credentials=context.server.verify_key,
            obj=peer_update,
        ).unwrap()

        return SyftSuccess(
            message=f"New route ({str(route)}) with id '{route.id}' "
            f"to peer {remote_server_peer.server_type.value} '{remote_server_peer.name}' "
            f"was added for {str(context.server.server_type)} '{context.server.name}'"
        )

    @service_method(
        path="network.delete_route_on_peer",
        name="delete_route_on_peer",
        unwrap_on_success=False,
    )
    def delete_route_on_peer(
        self,
        context: AuthedServiceContext,
        peer: ServerPeer,
        route: ServerRoute,
    ) -> SyftSuccess | SyftInfo:
        """
        Delete the route on the remote peer.

        Args:
            context (AuthedServiceContext): The authentication context for the service.
            peer (ServerPeer): The peer for which the route will be deleted.
            route (ServerRoute): The route to be deleted.

        Returns:
            SyftSuccess: If the route is successfully deleted.
            SyftInfo: If there is only one route left for the peer and
                the admin chose not to remove it
        """
        # creates a client on the remote server based on the credentials
        # of the current server's client
        remote_client = peer.client_with_context(context=context).unwrap()
        # ask the remote server to delete the route to the self server
        return remote_client.api.services.network.delete_route(
            peer_verify_key=context.credentials,
            route=route,
            called_by_peer=True,
        )

    @service_method(
        path="network.delete_route",
        name="delete_route",
        roles=GUEST_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def delete_route(
        self,
        context: AuthedServiceContext,
        peer_verify_key: SyftVerifyKey,
        route: ServerRoute | None = None,
        called_by_peer: bool = False,
    ) -> SyftSuccess | SyftInfo:
        """
        Delete a route for a given peer.
        If a peer has no routes left, there will be a prompt asking if the user want to remove it.
        If the answer is yes, it will be removed from the stash and will no longer be a peer.

        Args:
            context (AuthedServiceContext): The authentication context for the service.
            peer_verify_key (SyftVerifyKey): The verify key of the remote server peer.
            route (ServerRoute): The route to be deleted.
            called_by_peer (bool): The flag to indicate that it's called by a remote peer.

        Returns:
            SyftSuccess: If the route is successfully deleted.
            SyftInfo: If there is only one route left for the peer and
                the admin chose not to remove it
        """
        if called_by_peer and peer_verify_key != context.credentials:
            # verify if the peer is truly the one sending the request to delete the route to itself
            raise SyftException(
                public_message=(
                    f"The {type(peer_verify_key).__name__}: "
                    f"{peer_verify_key} does not match the signature of the message"
                )
            )
        remote_server_peer: ServerPeer = self._get_remote_server_peer_by_verify_key(
            context=context, peer_verify_key=peer_verify_key
        ).unwrap()

        if len(remote_server_peer.server_routes) == 1:
            warning_message = (
                f"There is only one route left to peer "
                f"{remote_server_peer.server_type.value} '{remote_server_peer.name}'. "
                f"Removing this route will remove the peer for "
                f"{str(context.server.server_type)} '{context.server.name}'."
            )
            response: bool = prompt_warning_message(
                message=warning_message,
                confirm=False,
            )
            if not response:
                return SyftInfo(
                    message=f"The last route to {remote_server_peer.server_type.value} "
                    f"'{remote_server_peer.name}' with id "
                    f"'{remote_server_peer.server_routes[0].id}' was not deleted."
                )

        if route is not None:
            remote_server_peer.delete_route(route)

        return_message = (
            f"Route '{str(route)}' to peer "
            f"{remote_server_peer.server_type.value} '{remote_server_peer.name}' "
            f"was deleted for {str(context.server.server_type)} '{context.server.name}'."
        )

        if len(remote_server_peer.server_routes) == 0:
            # remove the peer
            # TODO: should we do this as we are deleting the peer with a guest role level?
            self.stash.delete_by_uid(
                credentials=context.server.verify_key, uid=remote_server_peer.id
            ).unwrap()
            return_message += (
                f" There is no routes left to connect to peer "
                f"{remote_server_peer.server_type.value} '{remote_server_peer.name}', so it is deleted for "
                f"{str(context.server.server_type)} '{context.server.name}'."
            )
        else:
            # update the peer with the route removed
            peer_update = ServerPeerUpdate(
                id=remote_server_peer.id, server_routes=remote_server_peer.server_routes
            )
            self.stash.update(
                credentials=context.server.verify_key, obj=peer_update
            ).unwrap()

        return SyftSuccess(message=return_message)

    @service_method(
        path="network.update_route_priority_on_peer",
        name="update_route_priority_on_peer",
        unwrap_on_success=False,
    )
    def update_route_priority_on_peer(
        self,
        context: AuthedServiceContext,
        peer: ServerPeer,
        route: ServerRoute,
        priority: int | None = None,
    ) -> SyftSuccess:
        """
        Update the route priority on the remote peer.

        Args:
            context (AuthedServiceContext): The authentication context.
            peer (ServerPeer): The peer representing the remote server.
            route (ServerRoute): The route to be added.
            priority (int | None): The new priority value for the route. If not
                provided, it will be assigned the highest priority among all peers

        Returns:
            SyftSuccess: A success message if the route is verified,
                otherwise an error message.
        """
        # creates a client on the remote server based on the credentials
        # of the current server's client
        remote_client = peer.client_with_context(context=context).unwrap()
        result = remote_client.api.services.network.update_route_priority(
            peer_verify_key=context.credentials,
            route=route,
            priority=priority,
            called_by_peer=True,
        )
        return result

    @service_method(
        path="network.update_route_priority",
        name="update_route_priority",
        roles=GUEST_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def update_route_priority(
        self,
        context: AuthedServiceContext,
        peer_verify_key: SyftVerifyKey,
        route: ServerRoute,
        priority: int | None = None,
        called_by_peer: bool = False,
    ) -> SyftSuccess:
        """
        Updates a route's priority for the given peer

        Args:
            context (AuthedServiceContext): The authentication context for the service.
            peer_verify_key (SyftVerifyKey): The verify key of the peer whose route priority needs to be updated.
            route (ServerRoute): The route for which the priority needs to be updated.
            priority (int | None): The new priority value for the route. If not
                provided, it will be assigned the highest priority among all peers

        Returns:
            SyftSuccess : Successful response
        """
        if called_by_peer and peer_verify_key != context.credentials:
            raise SyftException(
                public_message=(
                    f"The {type(peer_verify_key).__name__}: "
                    f"{peer_verify_key} does not match the signature of the message"
                )
            )
        # get the full peer object from the store to update its routes
        remote_server_peer: ServerPeer = (
            self._get_remote_server_peer_by_verify_key(context, peer_verify_key)
        ).unwrap()
        # update the route's priority for the peer
        updated_server_route: ServerRouteType = (
            remote_server_peer.update_existed_route_priority(
                route=route, priority=priority
            ).unwrap()
        )
        new_priority: int = updated_server_route.priority
        # update the peer in the store
        peer_update = ServerPeerUpdate(
            id=remote_server_peer.id, server_routes=remote_server_peer.server_routes
        )
        self.stash.update(context.server.verify_key, peer_update).unwrap()
        return SyftSuccess(
            message=f"Route {route.id}'s priority updated to "
            f"{new_priority} for peer {remote_server_peer.name}"
        )

    @as_result(SyftException)
    def _get_remote_server_peer_by_verify_key(
        self, context: AuthedServiceContext, peer_verify_key: SyftVerifyKey
    ) -> ServerPeer:
        """
        Helper function to get the full server peer object from t
        he stash using its verify key
        """
        return self.stash.get_by_verify_key(
            credentials=context.server.verify_key,
            verify_key=peer_verify_key,
        ).unwrap()

    def _get_association_requests_by_peer_id(
        self, context: AuthedServiceContext, peer_id: UID
    ) -> list[Request]:
        """
        Get all the association requests from a peer. The association requests are sorted by request_time.
        """
        all_requests: list[Request] = context.server.services.request.get_all(context)
        association_requests: list[Request] = [
            request
            for request in all_requests
            if any(
                isinstance(change, AssociationRequestChange)
                and change.remote_peer.id == peer_id
                for change in request.changes
            )
        ]

        return sorted(
            association_requests, key=lambda request: request.request_time.utc_timestamp
        )


TYPE_TO_SERVICE[ServerPeer] = NetworkService
SERVICE_TO_TYPES[NetworkService].update({ServerPeer})


def from_server_url(context: TransformContext) -> TransformContext:
    if context.obj is not None and context.output is not None:
        url = context.obj.url.as_container_host()
        context.output["host_or_ip"] = url.host_or_ip
        context.output["protocol"] = url.protocol
        context.output["port"] = url.port
        context.output["private"] = False
        context.output["proxy_target_uid"] = context.obj.proxy_target_uid
        context.output["priority"] = 1
        context.output["rtunnel_token"] = context.obj.rtunnel_token

    return context


@transform(HTTPConnection, HTTPServerRoute)
def http_connection_to_server_route() -> list[Callable]:
    return [from_server_url]


def get_python_server_route(context: TransformContext) -> TransformContext:
    if context.output is not None and context.obj is not None:
        context.output["id"] = context.obj.server.id
        context.output["worker_settings"] = WorkerSettings.from_server(
            context.obj.server
        )
        context.output["proxy_target_uid"] = context.obj.proxy_target_uid
    return context


@transform(PythonConnection, PythonServerRoute)
def python_connection_to_server_route() -> list[Callable]:
    return [get_python_server_route]


@transform_method(PythonServerRoute, PythonConnection)
def server_route_to_python_connection(
    obj: Any, context: TransformContext | None = None
) -> list[Callable]:
    return PythonConnection(server=obj.server, proxy_target_uid=obj.proxy_target_uid)


@transform_method(HTTPServerRoute, HTTPConnection)
def server_route_to_http_connection(
    obj: Any, context: TransformContext | None = None
) -> list[Callable]:
    url = ServerURL(
        protocol=obj.protocol, host_or_ip=obj.host_or_ip, port=obj.port
    ).as_container_host()
    return HTTPConnection(
        url=url,
        proxy_target_uid=obj.proxy_target_uid,
        rtunnel_token=obj.rtunnel_token,
    )


@transform(ServerMetadata, ServerPeer)
def metadata_to_peer() -> list[Callable]:
    return [
        keep(
            [
                "id",
                "name",
                "verify_key",
                "server_type",
            ]
        ),
        make_set_default("admin_email", ""),
    ]


@transform(ServerSettings, ServerPeer)
def settings_to_peer() -> list[Callable]:
    return [
        keep(["id", "name", "verify_key", "server_type", "admin_email"]),
    ]



# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.result import as_result
from .api import TwinAPIEndpoint

MISSING_PATH_STRING = "Endpoint path: {path} does not exist."


@serializable(canonical_name="TwinAPIEndpointSQLStash", version=1)
class TwinAPIEndpointStash(ObjectStash[TwinAPIEndpoint]):
    @as_result(StashException, NotFoundException)
    def get_by_path(self, credentials: SyftVerifyKey, path: str) -> TwinAPIEndpoint:
        # TODO standardize by returning None if endpoint doesnt exist.
        res = self.get_one(
            credentials=credentials,
            filters={"path": path},
        )

        if res.is_err():
            raise NotFoundException(
                public_message=MISSING_PATH_STRING.format(path=path)
            )
        return res.unwrap()

    @as_result(StashException)
    def path_exists(self, credentials: SyftVerifyKey, path: str) -> bool:
        try:
            self.get_by_path(credentials=credentials, path=path).unwrap()
            return True
        except NotFoundException:
            return False



# stdlib
from collections.abc import Callable
from enum import Enum
from typing import Any

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...server.worker_settings import WorkerSettings
from ...server.worker_settings import WorkerSettingsV1
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...store.linked_obj import LinkedObject
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.syft_migration import migrate
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SYFT_OBJECT_VERSION_2
from ...types.syft_object import SyftObject
from ...types.transforms import TransformContext
from ...types.uid import UID
from ..action.action_permissions import ActionObjectPermission

__all__ = ["QueueItem"]


@serializable(canonical_name="Status", version=1)
class Status(str, Enum):
    CREATED = "created"
    PROCESSING = "processing"
    ERRORED = "errored"
    COMPLETED = "completed"
    INTERRUPTED = "interrupted"


@serializable()
class QueueItemV1(SyftObject):
    __canonical_name__ = "QueueItem"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_searchable__ = ["status", "worker_pool"]

    id: UID
    server_uid: UID
    result: Any | None = None
    resolved: bool = False
    status: Status = Status.CREATED

    method: str
    service: str
    args: list
    kwargs: dict[str, Any]
    job_id: UID | None = None
    worker_settings: WorkerSettingsV1 | None = None
    has_execute_permissions: bool = False
    worker_pool: LinkedObject


@serializable()
class QueueItem(SyftObject):
    __canonical_name__ = "QueueItem"
    __version__ = SYFT_OBJECT_VERSION_2

    __attr_searchable__ = ["status", "worker_pool_id"]

    id: UID
    server_uid: UID
    result: Any | None = None
    resolved: bool = False
    status: Status = Status.CREATED

    method: str
    service: str
    args: list
    kwargs: dict[str, Any]
    job_id: UID | None = None
    worker_settings: WorkerSettings | None = None
    has_execute_permissions: bool = False
    worker_pool: LinkedObject

    def __repr__(self) -> str:
        return f"<QueueItem: {self.id}>: {self.status}"

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        return f"<QueueItem: {self.id}>: {self.status}"

    @property
    def worker_pool_id(self) -> UID:
        return self.worker_pool.object_uid

    @property
    def is_action(self) -> bool:
        return self.service_path == "Action" and self.method_name == "execute"

    @property
    def action(self) -> Any:
        if self.is_action:
            return self.kwargs["action"]
        raise SyftException(public_message="QueueItem not an Action")


@serializable()
class ActionQueueItem(QueueItem):
    __canonical_name__ = "ActionQueueItem"
    __version__ = SYFT_OBJECT_VERSION_2

    method: str = "execute"
    service: str = "actionservice"


@serializable()
class APIEndpointQueueItem(QueueItem):
    __canonical_name__ = "APIEndpointQueueItem"
    __version__ = SYFT_OBJECT_VERSION_2

    method: str
    service: str = "apiservice"


@serializable()
class ActionQueueItemV1(QueueItemV1):
    __canonical_name__ = "ActionQueueItem"
    __version__ = SYFT_OBJECT_VERSION_1

    method: str = "execute"
    service: str = "actionservice"


@serializable()
class APIEndpointQueueItemV1(QueueItemV1):
    __canonical_name__ = "APIEndpointQueueItem"
    __version__ = SYFT_OBJECT_VERSION_1

    method: str
    service: str = "apiservice"


@serializable(canonical_name="QueueSQLStash", version=1)
class QueueStash(ObjectStash[QueueItem]):
    # FIX: Check the return value for None. set_result is used extensively
    @as_result(StashException)
    def set_result(
        self,
        credentials: SyftVerifyKey,
        item: QueueItem,
        add_permissions: list[ActionObjectPermission] | None = None,
    ) -> QueueItem | None:
        if item.resolved:
            self.check_type(item, self.object_type).unwrap()
            return super().update(credentials, item, add_permissions).unwrap()
        # TODO: should we log this?
        return None

    @as_result(SyftException)
    def set_placeholder(
        self,
        credentials: SyftVerifyKey,
        item: QueueItem,
        add_permissions: list[ActionObjectPermission] | None = None,
    ) -> QueueItem:
        # 🟡 TODO 36: Needs distributed lock
        if not item.resolved:
            try:
                self.get_by_uid(credentials, item.id).unwrap()
            except NotFoundException:
                self.check_type(item, self.object_type)
                return super().set(credentials, item, add_permissions).unwrap()
        return item

    @as_result(StashException)
    def pop(self, credentials: SyftVerifyKey, uid: UID) -> QueueItem | None:
        try:
            item = self.get_by_uid(credentials=credentials, uid=uid).unwrap()
        except NotFoundException:
            # TODO: Handle NotfoundException in code?
            return None

        self.delete_by_uid(credentials=credentials, uid=uid).unwrap()
        return item

    @as_result(StashException)
    def pop_on_complete(self, credentials: SyftVerifyKey, uid: UID) -> QueueItem:
        queue_item = self.get_by_uid(credentials=credentials, uid=uid).unwrap()
        if queue_item.status == Status.COMPLETED:
            self.delete_by_uid(credentials=credentials, uid=uid)
        return queue_item

    @as_result(StashException)
    def get_by_status(
        self, credentials: SyftVerifyKey, status: Status
    ) -> list[QueueItem]:
        return self.get_all(
            credentials=credentials,
            filters={"status": status},
        ).unwrap()

    @as_result(StashException)
    def _get_by_worker_pool(
        self, credentials: SyftVerifyKey, worker_pool: LinkedObject
    ) -> list[QueueItem]:
        worker_pool_id = worker_pool.object_uid

        return self.get_all(
            credentials=credentials,
            filters={"worker_pool_id": worker_pool_id},
        ).unwrap()


def upgrade_worker_settings_for_queue(context: TransformContext) -> TransformContext:
    if context.output and context.output["worker_settings"] is not None:
        worker_settings_old: WorkerSettingsV1 | None = context.output["worker_settings"]
        if worker_settings_old is None:
            return context

        worker_settings = worker_settings_old.migrate_to(
            WorkerSettings.__version__, context=context.to_server_context()
        )
        context.output["worker_settings"] = worker_settings

    return context


@migrate(QueueItemV1, QueueItem)
def migrate_queue_item_from_v1_to_v2() -> list[Callable]:
    return [upgrade_worker_settings_for_queue]


@migrate(ActionQueueItemV1, ActionQueueItem)
def migrate_action_queue_item_v1_to_v2() -> list[Callable]:
    return [upgrade_worker_settings_for_queue]


@migrate(APIEndpointQueueItemV1, APIEndpointQueueItem)
def migrate_api_endpoint_queue_item_v1_to_v2() -> list[Callable]:
    return [upgrade_worker_settings_for_queue]



# stdlib
from datetime import datetime
from typing import TYPE_CHECKING
from typing import cast

# relative
from ...serde.serializable import serializable
from ...store.linked_obj import LinkedObject
from ..context import AuthedServiceContext

if TYPE_CHECKING:
    # relative
    from ..request.request import Request
    from .notifications import Notification


class EmailTemplate:
    @staticmethod
    def email_title(notification: "Notification", context: AuthedServiceContext) -> str:
        raise NotImplementedError(
            "Email Template subclasses must implement the email_title method."
        )

    @staticmethod
    def email_body(notification: "Notification", context: AuthedServiceContext) -> str:
        raise NotImplementedError(
            "Email Template subclasses must implement the email_body method."
        )

    @staticmethod
    def batched_email_title(
        notifications: list["Notification"], context: AuthedServiceContext
    ) -> str:
        raise NotImplementedError(
            "Email Template subclasses must implement the batched_email_title method."
        )

    @staticmethod
    def batched_email_body(
        notifications: list["Notification"], context: AuthedServiceContext
    ) -> str:
        raise NotImplementedError(
            "Email Template subclasses must implement the batched_email_body method."
        )


@serializable(canonical_name="FailedJobTemplate", version=1)
class FailedJobTemplate(EmailTemplate):
    @staticmethod
    def email_title(notification: "Notification", context: AuthedServiceContext) -> str:
        return "Job Failed Notification"

    @staticmethod
    def email_body(notification: "Notification", context: AuthedServiceContext) -> str:
        notification.linked_obj = cast(LinkedObject, notification.linked_obj)
        queueitem_obj = notification.linked_obj.resolve_with_context(
            context=context
        ).unwrap()

        worker_pool_obj = queueitem_obj.worker_pool.resolve_with_context(
            context=context
        ).unwrap()
        method = queueitem_obj.method
        if queueitem_obj.service == "apiservice":
            method = queueitem_obj.kwargs.pop("path", "")
            queueitem_obj.kwargs.pop("log_id")

        head = """
            <head>
              <meta charset="UTF-8">
              <meta name="viewport" content="width=device-width, initial-scale=1.0">
              <title>Job Failed Notification</title>
              <style>
                body {
                  font-family: Arial, sans-serif;
                  background-color: #f4f4f4;
                  margin: 0;
                  padding: 0;
                }
                .container {
                  width: 100%;
                  max-width: 600px;
                  margin: 0 auto;
                  background-color: #ffffff;
                  padding: 20px;
                  border-radius: 8px;
                  box-shadow: 0 2px 4px rgba(0, 0, 0, 0.1);
                }
                h1 {
                  font-size: 24px;
                  color: #333333;
                }
                p {
                  font-size: 16px;
                  color: #666666;
                  line-height: 1.5;
                }
                .card {
                  background-color: #f9f9f9;
                  padding: 15px;
                  border-radius: 8px;
                  box-shadow: 0 1px 3px rgba(0, 0, 0, 0.1);
                  margin-top: 20px;
                }
                .card h2 {
                  font-size: 18px;
                  color: #333333;
                  margin-bottom: 10px;
                }
                .card p {
                  font-size: 14px;
                  color: #555555;
                  margin: 5px 0;
                }
                .footer {
                  font-size: 12px;
                  color: #999999;
                  margin-top: 30px;
                }
              </style>
            </head>
        """
        body = f"""
            <body>
              <div class="container">
                <h1>Job Failed Notification</h1>
                <p>Hello,</p>
                <p>We regret to inform you that your function job has encountered an
                unexpected error and could not be completed successfully.</p>

                <div class="card">
                  <h2>Job Details</h2>
                  <p><strong>Job ID:</strong> {queueitem_obj.job_id}</p>
                  <p><strong>Worker Pool:</strong> {worker_pool_obj.name}</p>
                  <p><strong>Method:</strong> {method}</p>
                  <p><strong>Service:</strong> {queueitem_obj.service}</p>
                  <p><strong>Arguments (args):</strong> {queueitem_obj.args}</p>
                  <p><strong>Keyword Arguments (kwargs):</strong> {queueitem_obj.kwargs}</p>
                </div>

                <p class="footer">If you need assistance, feel free to reach out to our support team.</p>
              </div>
            </body>
        """
        return f"""<html>{head} {body}</html>"""


@serializable(canonical_name="PasswordResetTemplate", version=1)
class PasswordResetTemplate(EmailTemplate):
    @staticmethod
    def email_title(notification: "Notification", context: AuthedServiceContext) -> str:
        return "Password Reset Requested"

    @staticmethod
    def email_body(notification: "Notification", context: AuthedServiceContext) -> str:
        user_service = context.server.services.user
        admin_verify_key = user_service.root_verify_key
        user = user_service.stash.get_by_verify_key(
            credentials=admin_verify_key, verify_key=notification.to_user_verify_key
        ).unwrap()
        if not user:
            raise Exception("User not found!")

        user.reset_token = user_service.generate_new_password_reset_token(
            context.server.settings.pwd_token_config
        )
        user.reset_token_date = datetime.now()

        result = user_service.stash.update(
            credentials=context.credentials, obj=user, has_permission=True
        )
        if result.is_err():
            raise Exception("Couldn't update the user password")

        expiry_time = context.server.services.settings.get(
            context=context
        ).pwd_token_config.token_exp_min

        head = """<head>
            <style>
                body {
                    font-family: Arial, sans-serif;
                    background-color: #f4f4f4;
                    color: #333;
                    margin: 0;
                    padding: 0;
                }
                .container {
                    max-width: 600px;
                    margin: 50px auto;
                    background-color: #ffffff;
                    padding: 20px;
                    border-radius: 8px;
                    box-shadow: 0 2px 4px rgba(0, 0, 0, 0.1);
                }
                h1 {
                    color: #333;
                    text-align: center;
                }
                p {
                    font-size: 16px;
                    line-height: 1.5;
                }
                .button {
                    display: block;
                    width: 200px;
                    margin: 30px auto;
                    padding: 10px;
                    background-color: #007BFF;
                    color: #fff;
                    text-align: center;
                    text-decoration: none;
                    border-radius: 4px;
                }
                .footer {
                    text-align: center;
                    font-size: 12px;
                    color: #aaa;
                    margin-top: 20px;
                }
            </style>
        </head>"""
        body = f"""<body>
            <div class="container">
                <h1>Password Reset</h1>
                <p>We received a request to reset your password. Your new temporary token is:</p>
                <h1>{user.reset_token}</h1>
                <p> Use
                    <code style="color: #FF8C00;background-color: #f0f0f0;font-size: 12px;">
                        syft_client.reset_password(token='{user.reset_token}', new_password=*****)
                    </code>.
                to reset your password. This token is valid for {expiry_time} seconds only.</p>
                <p>If you didn't request a password reset, please ignore this email.</p>
            </div>
        </body>"""
        return f"""<html>{head} {body}</html>"""


@serializable(canonical_name="OnboardEmailTemplate", version=1)
class OnBoardEmailTemplate(EmailTemplate):
    @staticmethod
    def email_title(notification: "Notification", context: AuthedServiceContext) -> str:
        return f"Welcome to {context.server.name} server!"

    @staticmethod
    def email_body(notification: "Notification", context: AuthedServiceContext) -> str:
        user_service = context.server.services.user
        admin_verify_key = user_service.root_verify_key
        admin = user_service.get_by_verify_key(admin_verify_key).unwrap()
        admin_name = admin.name

        head = (
            f"""
        <head>
            <title>Welcome to {context.server.name}</title>
        """
            + """
            <style>
                body {
                    font-family: Arial, sans-serif;
                    background-color: #f4f4f4;
                    color: #333;
                    line-height: 1.6;
                }
                .container {
                    max-width: 600px;
                    margin: 20px auto;
                    padding: 20px;
                    background: #fff;
                }
                h1 {
                    color: #0056b3;
                }
                .feature {
                    background-color: #e7f1ff;
                    padding: 10px;
                    margin: 10px 0;
                    border-radius: 5px;
                }
                .footer {
                    text-align: center;
                    font-size: 14px;
                    color: #aaa;
                }
            </style>
        </head>
        """
        )

        body = f"""
        <body>
            <div class="container">
                <h1>Welcome to {context.server.name} server!</h1>
                <p>Hello,</p>
                <p>We're thrilled to have you on board and
                excited to help you get started with our powerful features:</p>

                <div class="feature">
                    <h3>Remote Data Science</h3>
                    <p>Access and analyze data from anywhere, using our comprehensive suite of data science tools.</p>
                </div>

                <div class="feature">
                    <h3>Remote Code Execution</h3>
                    <p>Execute code remotely on private data, ensuring flexibility and efficiency in your research.</p>
                </div>

                <!-- Add more features here if needed -->

                <p>Explore these features and much more within your account.
                If you have any questions or need assistance, don't hesitate to reach out.</p>

                <p>Cheers,</p>
                <p>{admin_name}</p>

                <div class="footer">
                    This is an automated message, please do not reply directly to this email. <br>
                    For assistance, please contact our support team.
                </div>
            </div>
        </body>
        """
        return f"""<html>{head} {body}</html>"""


@serializable(canonical_name="RequestEmailTemplate", version=1)
class RequestEmailTemplate(EmailTemplate):
    @staticmethod
    def email_title(notification: "Notification", context: AuthedServiceContext) -> str:
        notification.linked_obj = cast(LinkedObject, notification.linked_obj)
        request_obj = notification.linked_obj.resolve_with_context(
            context=context
        ).unwrap()
        return f"Datasite {context.server.name}: A New Request ({str(request_obj.id)[:4]}) has been received!"

    @staticmethod
    def email_body(notification: "Notification", context: AuthedServiceContext) -> str:
        notification.linked_obj = cast(LinkedObject, notification.linked_obj)
        request_obj = notification.linked_obj.resolve_with_context(
            context=context
        ).unwrap()

        request_id = request_obj.id
        request_name = request_obj.requesting_user_name
        request_email = request_obj.requesting_user_email
        request_time = request_obj.request_time
        request_status = request_obj.status.name  # fails in l0 check right now
        request_changes = ",".join(
            [change.__class__.__name__ for change in request_obj.changes]
        )

        head = """
        <head>
            <title>Access Request Notification</title>
            <style>
                body {
                    font-family: Arial, sans-serif;
                    background-color: #f4f4f4;
                    color: #333;
                    padding: 20px;
                }
                .container {
                    max-width: 600px;
                    margin: 0 auto;
                    background: #fff;
                    padding: 20px;
                    border-radius: 8px;
                    box-shadow: 0 0 10px rgba(0, 0, 0, 0.1);
                }
                .header {
                    font-size: 24px;
                    color: #333;
                    text-align: center;
                }
                .content {
                    font-size: 16px;
                    line-height: 1.6;
                }

                .request-card {
                    background-color: #ffffff;
                    border: 1px solid #ddd;
                    padding: 15px;
                    margin-top: 20px;
                    border-radius: 8px;
                    box-shadow: 0 2px 4px rgba(0,0,0,0.1);
                }

                .request-header {
                    font-size: 18px;
                    color: #333;
                    margin-bottom: 10px;
                    font-weight: bold;
                }

                .request-content {
                    font-size: 14px;
                    line-height: 1.5;
                    color: #555;
                }
                .badge {
                    padding: 4px 10px;
                    border-radius: 15px;
                    color: white;
                    font-weight: bold;
                    margin: 10px;
                    box-shadow: 0px 4px 6px rgba(0,0,0,0.1);
                    text-transform: uppercase;
                }

                .yellow {
                    background-color: #fdd835;
                }

                .green {
                    background-color: #6dbf67;
                }

                .red {
                    background-color: #f7786b;
                }

                .button {
                    display: block;
                    width: max-content;
                    background-color: #007bff;
                    color: white;
                    padding: 10px 20px;
                    text-align: center;
                    text-color: white;
                    border-radius: 5px;
                    text-decoration: none;
                    font-weight: bold;
                    margin: 20px auto;
                }
                .footer {
                    text-align: center;
                    font-size: 14px;
                    color: #aaa;
                }
            </style>
        </head>"""

        body = f"""
        <body>
            <div class="container">
                <div class="header">
                    Request Notification
                </div>
                <div class="content">
                    <p>Hello,</p>
                    <p>A new request has been submitted and requires your attention.
                    Please review the details below:</p>

                    <div class="request-card">
                        <div class="request-header">Request Details</div>
                        <div class="request-content">

                            <p><strong>ID:</strong> {request_id}</p>
                            <p>
                            <strong>Submitted By:</strong>
                            {request_name}<{request_email}>
                            </p>
                            <p><strong>Date:</strong> {request_time}</p>
                        <div style="display: flex"><p><strong>Status:</strong><div class="badge yellow">
                            {request_status}
                        </div></div>
                        <p><strong>Changes:</strong>
                            {request_changes}
                        </p>

                        <p>Use:<br />
                        <code style="color: #FF8C00;background-color: #f0f0f0;font-size: 12px;">
                            request = client.api.services.request.get_by_uid(uid=sy.UID("{request_id}"))
                        </code><br />
                            to get this specific request.
                        </p>

                        <p>Or you can view all requests with: <br />
                        <code style="color: #FF8C00;background-color: #f0f0f0;font-size: 12px;">
                            client.requests
                        </code>
                        </p>

                        </div>
                    </div>
                    <p>If you did not expect this request or have concerns about it,
                    please contact our support team immediately.</p>
                </div>
                <div class="footer">
                    This is an automated message, please do not reply directly to this email. <br>
                    For assistance, please contact our support team.
                </div>
            </div>
        </body>
        """
        return f"""<html>{head} {body}</html>"""

    @staticmethod
    def batched_email_title(
        notifications: list["Notification"], context: AuthedServiceContext
    ) -> str:
        return "Batched Requests Notifications"

    @staticmethod
    def batched_email_body(
        notifications: list["Notification"], context: AuthedServiceContext
    ) -> str:
        notifications_info = ""
        for i, notification in enumerate(notifications):
            if i > 3:
                break
            notification.linked_obj = cast(LinkedObject, notification.linked_obj)
            request_obj = notification.linked_obj.resolve_with_context(
                context=context
            ).unwrap()

            request_id = request_obj.id
            request_name = request_obj.requesting_user_name
            request_email = request_obj.requesting_user_email
            request_time = request_obj.request_time
            request_status = request_obj.status.name  # fails in l0 check right now
            request_changes = ",".join(
                [change.__class__.__name__ for change in request_obj.changes]
            )

            notifications_info += f"""<tr>
                                <td>{str(request_id)[:4] + "..."}</td>
                                <td>{request_name}</td>
                                <td>{request_email}</td>
                                <td>{request_time}</td>
                                <td>{request_status}</td>
                                <td>{request_changes}</td>
                            </tr>"""

        see_more_info = ""
        if len(notifications) > 4:
            see_more_info = f"""<p class="more-requests">{len(notifications) - 4}
                  more requests made during this time period.
                  Connect to the server to check all requests.</p>"""
        head = """
            <head>
              <meta charset="UTF-8">
              <meta name="viewport" content="width=device-width, initial-scale=1.0">
              <title>Batched Requests Notification</title>
              <style>
                body {
                  font-family: Arial, sans-serif;
                  background-color: #f4f4f4;
                  margin: 0;
                  padding: 0;
                  color: #333;
                }
                .container {
                  width: 100%;
                  max-width: 600px;
                  margin: 0 auto;
                  background-color: #ffffff;
                  padding: 20px;
                  border-radius: 8px;
                  box-sizing: border-box; /* Added to include padding in width */
                  box-shadow: 0 0 10px rgba(0, 0, 0, 0.1);
                }
                .header {
                  font-size: 18px;
                  font-weight: bold;
                  margin-bottom: 20px;
                  color: #4CAF50;
                }
                .content {
                  font-size: 14px;
                  line-height: 1.6;
                  color: #555555;
                }
                .content p {
                  margin: 10px 0;
                }
                .request-table {
                  width: 100%;
                  border-collapse: collapse;
                  margin-top: 20px;
                  table-layout: fixed; /* Added to fix table layout */
                }
                .request-table th, .request-table td {
                  text-align: left;
                  padding: 12px;
                  border: 1px solid #ddd;
                  word-wrap: break-word; /* Added to wrap long content */
                }
                .request-table th {
                  background-color: #f8f8f8;
                  color: #333;
                  font-weight: bold;
                }
                .request-table tr:nth-child(even) {
                  background-color: #f9f9f9;
                }
                .more-requests {
                  font-size: 13px;
                  color: #FF5722;
                  margin-top: 10px;
                }
                .footer {
                  margin-top: 20px;
                  font-size: 12px;
                  color: #777777;
                }
                .button {
                  background-color: #4CAF50;
                  color: #ffffff;
                  padding: 10px 20px;
                  text-decoration: none;
                  border-radius: 5px;
                  font-size: 14px;
                  display: inline-block;
                  margin-top: 20px;
                }
              </style>
            </head>
        """
        body = f"""
            <body>
              <div class="container">
                <div class="header">
                  Batched Requests Notification
                </div>
                <div class="content">
                  <p>Hello Admin,</p>
                  <p>This is to inform you that a batch of requests has been processed.
                  Below are the details of the most recent requests:</p>
                  <table class="request-table">
                    <thead>
                      <tr>
                        <th>Request ID</th>
                        <th>User</th>
                        <th>User Email</th>
                        <th>Date</th>
                        <th>Status</th>
                        <th>Changes</th>
                      </tr>
                    </thead>
                    <tbody>
                      {notifications_info}
                      <!-- Only show the first 3 requests -->
                    </tbody>
                  </table>
                  {see_more_info}
                </div>
                <div class="footer">
                  <p>Thank you,</p>
                </div>
              </div>
            </body>
        """
        return f"""<html>{head} {body}</html>"""


@serializable(canonical_name="RequestUpdateEmailTemplate", version=1)
class RequestUpdateEmailTemplate(EmailTemplate):
    @staticmethod
    def email_title(notification: "Notification", context: AuthedServiceContext) -> str:
        return f"Datasite {context.server.name}: {notification.subject}"

    @staticmethod
    def email_body(notification: "Notification", context: AuthedServiceContext) -> str:
        # relative
        from ..request.request import RequestStatus

        notification.linked_obj = cast(LinkedObject, notification.linked_obj)
        request_obj: Request = notification.linked_obj.resolve_with_context(
            context=context
        ).unwrap()
        badge_color = "red" if request_obj.status.name == "REJECTED" else "green"

        request_id = request_obj.id
        request_name = request_obj.requesting_user_name
        request_email = request_obj.requesting_user_email
        request_time = request_obj.request_time
        request_status = request_obj.status
        request_status_name = request_status.name  # fails in l0 check right now
        request_changes = ",".join(
            [change.__class__.__name__ for change in request_obj.changes]
        )

        deny_reason_html = ""
        if request_status == RequestStatus.REJECTED:
            deny_reason_or_err = request_obj.get_deny_reason(context=context)
            if deny_reason_or_err.is_err():
                deny_reason = None
            else:
                deny_reason = deny_reason_or_err.unwrap()

            if not isinstance(deny_reason, str) or not len(deny_reason):
                deny_reason = (
                    "No reason provided, please contact the admin for more information."
                )

            deny_reason_html = f"<p><strong>Deny Reason:</strong> {deny_reason}</p>"

        head = """
        <head>
            <title>Access Request Notification</title>
            <style>
                body {
                    font-family: Arial, sans-serif;
                    background-color: #f4f4f4;
                    color: #333;
                    padding: 20px;
                }
                .container {
                    max-width: 600px;
                    margin: 0 auto;
                    background: #fff;
                    padding: 20px;
                    border-radius: 8px;
                    box-shadow: 0 0 10px rgba(0, 0, 0, 0.1);
                }
                .header {
                    font-size: 24px;
                    color: #333;
                    text-align: center;
                }
                .content {
                    font-size: 16px;
                    line-height: 1.6;
                }

                .request-card {
                    background-color: #ffffff;
                    border: 1px solid #ddd;
                    padding: 15px;
                    margin-top: 20px;
                    border-radius: 8px;
                    box-shadow: 0 2px 4px rgba(0,0,0,0.1);
                }

                .request-header {
                    font-size: 18px;
                    color: #333;
                    margin-bottom: 10px;
                    font-weight: bold;
                }

                .badge {
                    padding: 4px 10px;
                    border-radius: 15px;
                    color: white;
                    font-weight: bold;
                    margin: 10px;
                    box-shadow: 0px 4px 6px rgba(0,0,0,0.1);
                    text-transform: uppercase;
                }

                .yellow {
                    background-color: #fdd835;
                }

                .green {
                    background-color: #6dbf67;
                }

                .red {
                    background-color: #f7786b;
                }

                .request-content {
                    font-size: 14px;
                    line-height: 1.5;
                    color: #555;
                }

                .button {
                    display: block;
                    width: max-content;
                    background-color: #007bff;
                    color: white;
                    padding: 10px 20px;
                    text-align: center;
                    text-color: white;
                    border-radius: 5px;
                    text-decoration: none;
                    font-weight: bold;
                    margin: 20px auto;
                }
                .footer {
                    text-align: center;
                    font-size: 14px;
                    color: #aaa;
                }
            </style>
        </head>"""

        body = f"""
        <body>
            <div class="container">
                <div class="header">
                    Request Notification
                </div>
                <div class="content">
                    <p>Hello,</p>
                    <p>The status of your recent request has been updated.
                    Below is the latest information regarding it:</p>

                    <div class="request-card">
                        <div class="request-header">Request Details</div>
                        <div class="request-content">

                            <p><strong>ID:</strong> {request_id}</p>
                            <p>
                            <strong>Submitted By:</strong>
                            {request_name} {request_email}
                            </p>
                            <p><strong>Date:</strong> {request_time}</p>
                            <div style="display: flex"><p><strong>Status:</strong><div class="badge {badge_color}">
                                {request_status_name}
                            </div>
                            </div>
                            {deny_reason_html}
                            <p>
                            <strong>Changes:</strong>
                            {request_changes}
                            </p>

                            <p>Use:<br />
                            <code style="color: #FF8C00;background-color: #f0f0f0;font-size: 12px;">
                                request = client.api.services.request.get_by_uid(uid=sy.UID("{request_id}"))
                            </code><br />
                                to get this specific request.
                            </p>

                            <p>Or you can view all requests with: <br />
                            <code style="color: #FF8C00;background-color: #f0f0f0;font-size: 12px;">
                                client.requests
                            </code>
                            </p>
                        </div>
                    </div>
                    <p>If you did not expect this request or have concerns about it,
                    please contact our support team immediately.</p>
                </div>
                <div class="footer">
                    This is an automated message, please do not reply directly to this email. <br>
                    For assistance, please contact our support team.
                </div>
            </div>
        </body>
        """
        return f"""<html>{head} {body}</html>"""



# stdlib
from collections.abc import Callable

# syft absolute
import syft as sy

# relative
from ..... import test_settings
from ..rate_limiter import is_within_rate_limit


def make_schema(settings: dict, worker_pool_name: str) -> Callable:
    updated_settings = {
        "calls_per_min": 5,
        "rate_limiter_enabled": True,
        "credentials": test_settings.gce_service_account.to_dict(),
        "region": test_settings.gce_region,
        "project_id": test_settings.gce_project_id,
        "dataset_1": test_settings.dataset_1,
        "table_1": test_settings.table_1,
        "table_2": test_settings.table_2,
    } | settings

    @sy.api_endpoint(
        path="bigquery.schema",
        description="This endpoint allows for visualising the metadata of tables available in BigQuery.",
        settings=updated_settings,
        helper_functions=[
            is_within_rate_limit
        ],  # Adds ratelimit as this is also a method available to data scientists
        worker_pool_name=worker_pool_name,
    )
    def live_schema(
        context,
    ) -> str:
        # stdlib
        import datetime

        # third party
        from google.cloud import bigquery  # noqa: F811
        from google.oauth2 import service_account
        import pandas as pd

        # syft absolute
        from syft import SyftException

        # Auth for Bigquer based on the workload identity
        credentials = service_account.Credentials.from_service_account_info(
            context.settings["credentials"]
        )
        scoped_credentials = credentials.with_scopes(
            ["https://www.googleapis.com/auth/cloud-platform"]
        )

        client = bigquery.Client(
            credentials=scoped_credentials,
            location=context.settings["region"],
        )

        # Store a dict with the calltimes for each user, via the email.
        if context.settings["rate_limiter_enabled"]:
            if context.user.email not in context.state.keys():
                context.state[context.user.email] = []

            if not context.code.is_within_rate_limit(context):
                raise SyftException(
                    public_message="Rate limit of calls per minute has been reached."
                )
            context.state[context.user.email].append(datetime.datetime.now())

        try:
            # Formats the data schema in a data frame format
            # Warning: the only supported format types are primitives, np.ndarrays and pd.DataFrames

            data_schema = []
            for table_id in [
                f"{context.settings['dataset_1']}.{context.settings['table_1']}",
                f"{context.settings['dataset_1']}.{context.settings['table_2']}",
            ]:
                table = client.get_table(table_id)
                for schema in table.schema:
                    data_schema.append(
                        {
                            "project": str(table.project),
                            "dataset_id": str(table.dataset_id),
                            "table_id": str(table.table_id),
                            "schema_name": str(schema.name),
                            "schema_field": str(schema.field_type),
                            "description": str(table.description),
                            "num_rows": str(table.num_rows),
                        }
                    )
            return pd.DataFrame(data_schema)

        except Exception as e:
            # not a bigquery exception
            if not hasattr(e, "_errors"):
                output = f"got exception e: {type(e)} {str(e)}"
                raise SyftException(
                    public_message=f"An error occured executing the API call {output}"
                )

            # Should add appropriate error handling for what should be exposed to the data scientists.
            raise SyftException(
                public_message="An error occured executing the API call, please contact the domain owner."
            )

    return live_schema



# stdlib

# third party

# third party
from sqlalchemy.orm import Session

# relative
from ...custom_worker.config import DockerWorkerConfig
from ...custom_worker.config import WorkerConfig
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.db.stash import with_session
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.errors import SyftException
from ...types.result import as_result
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from .worker_image import SyftWorkerImage


@serializable(canonical_name="SyftWorkerImageSQLStash", version=1)
class SyftWorkerImageStash(ObjectStash[SyftWorkerImage]):
    @as_result(SyftException, StashException, NotFoundException)
    @with_session
    def set(
        self,
        credentials: SyftVerifyKey,
        obj: SyftWorkerImage,
        add_permissions: list[ActionObjectPermission] | None = None,
        add_storage_permission: bool = True,
        ignore_duplicates: bool = False,
        session: Session = None,
        skip_check_type: bool = False,
    ) -> SyftWorkerImage:
        # By default syft images have all read permission
        add_permissions = [] if add_permissions is None else add_permissions
        add_permissions.append(
            ActionObjectPermission(uid=obj.id, permission=ActionPermission.ALL_READ)
        )

        if isinstance(obj.config, DockerWorkerConfig):
            worker_config_exists = self.worker_config_exists(
                credentials=credentials, config=obj.config
            ).unwrap()
            if worker_config_exists:
                raise SyftException(
                    public_message=f"Worker Image with config {obj.config} already exists"
                )

        return (
            super()
            .set(
                credentials,
                obj,
                add_permissions=add_permissions,
                add_storage_permission=add_storage_permission,
                ignore_duplicates=ignore_duplicates,
                session=session,
            )
            .unwrap()
        )

    @as_result(StashException, NotFoundException)
    def worker_config_exists(
        self, credentials: SyftVerifyKey, config: WorkerConfig
    ) -> bool:
        try:
            self.get_by_worker_config(credentials=credentials, config=config).unwrap()
            return True
        except NotFoundException:
            return False

    @as_result(StashException, NotFoundException)
    def get_by_worker_config(
        self, credentials: SyftVerifyKey, config: WorkerConfig
    ) -> SyftWorkerImage:
        # TODO cannot search on fields containing objects
        all_images = self.get_all(credentials=credentials).unwrap()
        for image in all_images:
            if image.config == config:
                return image
        raise NotFoundException(
            public_message=f"Worker Image with config {config} not found"
        )



# third party
from pydantic import BaseModel

# relative
from ..serde.serializable import serializable


@serializable(canonical_name="LockingConfig", version=1)
class LockingConfig(BaseModel):
    lock_name: str = "syft_lock"
    namespace: str | None = None
    expire: int | None = 60
    timeout: int | None = 30
    retry_interval: float = 0.1


@serializable(canonical_name="NoLockingConfig", version=1)
class NoLockingConfig(LockingConfig):
    pass


@serializable(canonical_name="ThreadingLockingConfig", version=1)
class ThreadingLockingConfig(LockingConfig):
    pass



# stdlib

# stdlib

# relative
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...store.linked_obj import LinkedObject
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..notification.notifications import CreateNotification
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..user.user_roles import ServiceRole
from .project import Project
from .project import ProjectEvent
from .project import ProjectRequest
from .project import ProjectSubmit
from .project import create_project_hash
from .project_stash import ProjectStash


@serializable(canonical_name="ProjectService", version=1)
class ProjectService(AbstractService):
    stash: ProjectStash

    def __init__(self, store: DBManager) -> None:
        self.stash = ProjectStash(store=store)

    @as_result(SyftException)
    def validate_project_leader(
        self, context: AuthedServiceContext, project: Project
    ) -> None:
        if project.state_sync_leader.verify_key != context.server.verify_key:
            error_msg = "Only the project leader can do this operation"
            raise SyftException(public_message=error_msg)

    @as_result(SyftException)
    def validate_user_permission_for_project(
        self, context: AuthedServiceContext, project: Project
    ) -> None:
        if not project.has_permission(context.credentials):
            error_msg = "User does not have permission to sync events"
            raise SyftException(public_message=error_msg)

    @as_result(StashException)
    def project_exists(
        self, context: AuthedServiceContext, project: ProjectSubmit
    ) -> bool:
        credentials = context.server.verify_key
        try:
            self.stash.get_by_uid(credentials=credentials, uid=project.id).unwrap()
            return True
        except NotFoundException:
            return False

    @as_result(SyftException)
    def validate_project_event_seq(
        self, project_event: ProjectEvent, project: Project
    ) -> None:
        if project_event.seq_no is None:
            raise SyftException(public_message=f"{project_event}.seq_no is None")
        if project_event.seq_no <= len(project.events) and len(project.events) > 0:
            # TODO: We need a way to handle alert returns...
            # e.g. here used to be:
            # SyftNotReady(message="Project out of sync event")
            raise SyftException(public_message="Project events are out of sync")
        if project_event.seq_no > len(project.events) + 1:
            raise SyftException(public_message="Project events are out of order")

    def is_project_leader(
        self, context: AuthedServiceContext, project: Project
    ) -> bool:
        return context.credentials == project.state_sync_leader.verify_key

    @service_method(
        path="project.can_create_project",
        name="can_create_project",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def can_create_project(self, context: AuthedServiceContext) -> bool:
        role = context.server.services.user.get_role_for_credentials(
            credentials=context.credentials
        ).unwrap()

        if role >= ServiceRole.DATA_SCIENTIST:
            return True

        raise SyftException(
            public_message="You do not have permission to create projects. Contact your admin."
        )

    @service_method(
        path="project.create_project",
        name="create_project",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def create_project(
        self, context: AuthedServiceContext, project: ProjectSubmit
    ) -> SyftSuccess:
        """Start a Project"""
        self.can_create_project(context)

        project_exists = self.project_exists(context, project).unwrap()
        if project_exists:
            raise SyftException(public_message=f"Project {project.id} already exists")

        _project: Project = project.to(Project, context=context)

        # Updating the leader server route of the project object
        # In case the current server, is the leader, they would input their server route
        # For the followers, they would check if the leader is their server peer
        # using the leader's verify_key
        # If the follower do not have the leader as its peer in its routes
        # They would raise as error
        leader_server = _project.state_sync_leader

        # If the current server is a follower
        # For followers the leader server route is retrieved from its peer
        if leader_server.verify_key != context.server.verify_key:
            # FIX: networkservice stash to new BaseStash
            peer_id = context.server.id.short() if context.server.id else ""
            leader_server_peer = (
                context.server.services.network.stash.get_by_verify_key(
                    credentials=context.server.verify_key,
                    verify_key=leader_server.verify_key,
                ).unwrap(
                    public_message=(
                        f"Leader Server(id={leader_server.id.short()}) is not a "
                        f"peer of this Server(id={peer_id})"
                    )
                )
            )
        else:
            # for the leader server, as it does not have route information to itself
            # we rely on the data scientist to provide the route
            # the route is then validated by the leader
            if project.leader_server_route is not None:
                leader_server_peer = project.leader_server_route.validate_with_context(
                    context=context
                ).unwrap()
            else:
                raise SyftException(
                    public_message=f"project {project}'s leader_server_route is None"
                )

        _project.leader_server_peer = leader_server_peer
        # This should always be the last call before flushing to DB
        _project.start_hash = create_project_hash(_project)[1]

        stored_project: Project = self.stash.set(context.credentials, _project).unwrap()
        stored_project = self.add_signing_key_to_project(context, stored_project)

        return SyftSuccess(message="Project successfully created", value=stored_project)

    @service_method(
        path="project.add_event",
        name="add_event",
        roles=GUEST_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def add_event(
        self, context: AuthedServiceContext, project_event: ProjectEvent
    ) -> SyftSuccess:
        """Add events to a project"""
        # Event object should be received from the leader of the project
        if not isinstance(project_event, ProjectEvent):
            raise SyftException(
                public_message="project_event should be a ProjectEvent object"
            )

        project = self.stash.get_by_uid(
            context.server.verify_key, uid=project_event.project_id
        ).unwrap()

        # FIX: MERGE: Rename function below
        self.validate_project_leader(context, project).unwrap(
            public_message="Project Events should be passed to leader by broadcast endpoint"
        )

        if not self.is_project_leader(context, project):
            raise SyftException(
                public_message="Only the leader of the project can add events"
            )

        project.events.append(project_event)
        project.event_id_hashmap[project_event.id] = project_event

        # TODO: better name for the function should be check_and_notify or something?
        self.check_for_project_request(project, project_event, context)

        updated_project = self.stash.update(context.server.verify_key, project).unwrap()

        return SyftSuccess(
            message=f"Project event {project_event.id} added successfully",
            value=updated_project,
        )

    @service_method(
        path="project.broadcast_event",
        name="broadcast_event",
        roles=GUEST_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def broadcast_event(
        self, context: AuthedServiceContext, project_event: ProjectEvent
    ) -> SyftSuccess:
        """To add events to a projects"""
        # Only the leader of the project could add events to the projects
        # Any Event to be added to the project should be sent to the leader of the project
        # The leader broadcasts the event to all the members of the project

        project = self.stash.get_by_uid(
            context.server.verify_key, uid=project_event.project_id
        ).unwrap()

        self.validate_project_leader(context, project).unwrap(
            public_message="Only the leader of the project can broadcast events"
        )
        self.validate_user_permission_for_project(context, project)
        self.validate_project_event_seq(project_event, project).unwrap()

        project.events.append(project_event)
        project.event_id_hashmap[project_event.id] = project_event

        self.check_for_project_request(project, project_event, context)

        # Broadcast the event to all the members of the project
        for member in project.members:
            if member.verify_key != context.server.verify_key:
                # Retrieving the ServerPeer Object to communicate with the server
                peer = context.server.services.network.stash.get_by_verify_key(
                    credentials=context.server.verify_key,
                    verify_key=member.verify_key,
                ).unwrap(
                    public_message=f"Leader server does not have peer {member.name}-{member.id.short()}"
                    + ". Please exchange routes with the peer."
                )
                remote_client = peer.client_with_context(context=context).unwrap(
                    public_message=f"Failed to create remote client for peer: {peer.id}."
                )
                remote_client.api.services.project.add_event(project_event)

        updated_project = self.stash.update(context.server.verify_key, project).unwrap()

        return SyftSuccess(
            message=f"Event #{project_event.seq_no} of {project.name} broadcasted successfully",
            value=updated_project,
        )

    @service_method(
        path="project.sync",
        name="sync",
        roles=GUEST_ROLE_LEVEL,
    )
    def sync(
        self, context: AuthedServiceContext, project_id: UID, seq_no: int
    ) -> list[ProjectEvent]:
        """Given a starting event seq_no, gets all following events from a project"""
        if seq_no < 0:
            raise SyftException(
                public_message="Input seq_no should be a non negative integer"
            )

        # Event object should be received from the leader of the project
        project = self.stash.get_by_uid(
            context.server.verify_key, uid=project_id
        ).unwrap()

        self.validate_project_leader(context, project)
        self.validate_user_permission_for_project(context, project)

        return project.events[seq_no:]

    @service_method(path="project.get_all", name="get_all", roles=GUEST_ROLE_LEVEL)
    def get_all(self, context: AuthedServiceContext) -> list[Project]:
        projects: list[Project] = self.stash.get_all(context.credentials).unwrap()

        for idx, project in enumerate(projects):
            projects[idx] = self.add_signing_key_to_project(context, project)

        return projects

    @service_method(
        path="project.get_by_name",
        name="get_by_name",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_by_name(self, context: AuthedServiceContext, name: str) -> Project:
        try:
            project = self.stash.get_by_name(
                context.credentials, project_name=name
            ).unwrap()
        except NotFoundException as exc:
            raise NotFoundException.from_exception(
                exc, public_message="Project '{name}' does not exist"
            )

        return self.add_signing_key_to_project(context, project)

    @service_method(
        path="project.get_by_uid",
        name="get_by_uid",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_by_uid(self, context: AuthedServiceContext, uid: UID) -> Project:
        try:
            credentials = context.server.verify_key
            return self.stash.get_by_uid(credentials=credentials, uid=uid).unwrap()
        except NotFoundException as exc:
            raise NotFoundException.from_exception(
                exc, public_message=f"Project {uid} not found"
            )

    as_result(StashException, NotFoundException)

    def add_signing_key_to_project(
        self, context: AuthedServiceContext, project: Project
    ) -> Project:
        try:
            user = context.server.services.user.stash.get_by_verify_key(
                credentials=context.credentials, verify_key=context.credentials
            ).unwrap()
        except NotFoundException as exc:
            raise NotFoundException.from_exception(
                exc, public_message="User not found! Please register the user first"
            )
        # Automatically infuse signing key of user
        project.user_signing_key = user.signing_key

        return project

    # TODO: Glob Notification error here
    @as_result(SyftException)
    def check_for_project_request(
        self,
        project: Project,
        project_event: ProjectEvent,
        context: AuthedServiceContext,
    ) -> None:
        # TODO: Should we really raise an exception if notification fails to be sent?
        #       Maybe logging and moving on is better?
        """
        Checks if there are any ProjectEvent requests and messages the admin
        in case there is one.

        This method raises an exception if the notification fails to send.

        Args:
            project (Project): Project object
            project_event (ProjectEvent): Project event object
            context (AuthedServiceContext): Context of the server

        Returns:
            SyftSuccess: SyftSuccess if message is created else SyftError
        """
        if (
            isinstance(project_event, ProjectRequest)
            and project_event.linked_request.server_uid == context.server.id
        ):
            link = LinkedObject.with_context(project, context=context)

            message = CreateNotification(
                subject=f"A new request has been added to the project {project.name}",
                from_user_verify_key=context.credentials,
                to_user_verify_key=context.server.verify_key,
                linked_obj=link,
            )

            # TODO: Update noteificationservice result
            result = context.server.services.notification.send(
                context=context, notification=message
            )
            if isinstance(result, SyftError):
                raise SyftException(public_message=result)


TYPE_TO_SERVICE[Project] = ProjectService
SERVICE_TO_TYPES[ProjectService].update({Project})



# stdlib
import importlib
import logging
from typing import Any

# third party
import numpy as np

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.db import DBManager
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.datetime import DateTime
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.syft_object import SyftObject
from ...types.twin_object import TwinObject
from ...types.uid import UID
from ..code.user_code import UserCode
from ..code.user_code import execute_byte_code
from ..context import AuthedServiceContext
from ..policy.policy import OutputPolicy
from ..policy.policy import retrieve_from_db
from ..response import SyftResponseMessage
from ..response import SyftSuccess
from ..response import SyftWarning
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import UserLibConfigRegistry
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..user.user_roles import ServiceRole
from .action_endpoint import CustomEndpointActionObject
from .action_object import Action
from .action_object import ActionObject
from .action_object import ActionObjectPointer
from .action_object import ActionType
from .action_object import AnyActionObject
from .action_object import TwinMode
from .action_permissions import ActionObjectPermission
from .action_permissions import ActionObjectREAD
from .action_permissions import ActionPermission
from .action_permissions import StoragePermission
from .action_store import ActionObjectStash
from .action_types import action_type_for_type
from .numpy import NumpyArrayObject
from .pandas import PandasDataFrameObject  # noqa: F401
from .pandas import PandasSeriesObject  # noqa: F401

logger = logging.getLogger(__name__)


@serializable(canonical_name="ActionService", version=1)
class ActionService(AbstractService):
    stash: ActionObjectStash

    def __init__(self, store: DBManager) -> None:
        self.stash = ActionObjectStash(store)

    @service_method(path="action.np_array", name="np_array")
    def np_array(self, context: AuthedServiceContext, data: Any) -> Any:
        # TODO: REMOVE!
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        # cast here since we are sure that AuthedServiceContext has a server

        np_obj = NumpyArrayObject(
            dtype=data.dtype,
            shape=data.shape,
            syft_action_data_cache=data,
            syft_server_location=context.server.id,
            syft_client_verify_key=context.credentials,
        )
        blob_store_result = np_obj._save_to_blob_storage().unwrap()
        if isinstance(blob_store_result, SyftWarning):
            logger.debug(blob_store_result.message)

        np_pointer = self._set(context, np_obj)
        return np_pointer

    @service_method(
        path="action.set",
        name="set",
        roles=GUEST_ROLE_LEVEL,
    )
    def set(
        self,
        context: AuthedServiceContext,
        action_object: ActionObject | TwinObject,
        add_storage_permission: bool = True,
        ignore_detached_objs: bool = False,
    ) -> ActionObject:
        return self._set(
            context,
            action_object,
            has_result_read_permission=True,
            add_storage_permission=add_storage_permission,
            ignore_detached_objs=ignore_detached_objs,
        ).unwrap()

    def is_detached_obj(
        self,
        action_object: ActionObject | TwinObject,
        ignore_detached_obj: bool = False,
    ) -> bool:
        """
        A detached object is an object that is not yet saved to the blob storage.
        """
        if (
            isinstance(action_object, TwinObject)
            and (
                (
                    action_object.mock_obj.syft_action_saved_to_blob_store
                    and action_object.mock_obj.syft_blob_storage_entry_id is None
                )
                or (
                    action_object.private_obj.syft_action_saved_to_blob_store
                    and action_object.private_obj.syft_blob_storage_entry_id is None
                )
            )
            and not ignore_detached_obj
        ):
            return True
        if isinstance(action_object, ActionObject) and (
            action_object.syft_action_saved_to_blob_store
            and action_object.syft_blob_storage_entry_id is None
            and not ignore_detached_obj
        ):
            return True
        return False

    @as_result(StashException, SyftException)
    def _set(
        self,
        context: AuthedServiceContext,
        action_object: ActionObject | TwinObject,
        has_result_read_permission: bool = False,
        add_storage_permission: bool = True,
        ignore_detached_objs: bool = False,
    ) -> ActionObject:
        if self.is_detached_obj(action_object, ignore_detached_objs):
            raise SyftException(
                public_message="You uploaded an ActionObject that is not yet in the blob storage"
            )

        """Save an object to the action store"""
        # 🟡 TODO 9: Create some kind of type checking / protocol for SyftSerializable

        if isinstance(action_object, ActionObject):
            action_object.syft_created_at = DateTime.now()
            (
                action_object._clear_cache()
                if action_object.syft_action_saved_to_blob_store
                else None
            )
        else:  # TwinObject
            action_object.private_obj.syft_created_at = DateTime.now()  # type: ignore[unreachable]
            action_object.mock_obj.syft_created_at = DateTime.now()

            # Clear cache if data is saved to blob storage
            (
                action_object.private_obj._clear_cache()
                if action_object.private_obj.syft_action_saved_to_blob_store
                else None
            )
            (
                action_object.mock_obj._clear_cache()
                if action_object.mock_obj.syft_action_saved_to_blob_store
                else None
            )

        # If either context or argument is True, has_result_read_permission is True
        has_result_read_permission = (
            context.extra_kwargs.get("has_result_read_permission", False)
            or has_result_read_permission
        )

        self.stash.set_or_update(
            uid=action_object.id,
            credentials=context.credentials,
            syft_object=action_object,
            has_result_read_permission=has_result_read_permission,
            add_storage_permission=add_storage_permission,
        ).unwrap()

        if isinstance(action_object, TwinObject):
            # give read permission to the mock
            # if mock is saved to blob store, then add READ permission
            if action_object.mock_obj.syft_action_saved_to_blob_store:
                blob_id = action_object.mock_obj.syft_blob_storage_entry_id
                permission = ActionObjectPermission(blob_id, ActionPermission.ALL_READ)
                # add_permission is not resultified.
                context.server.services.blob_storage.stash.add_permission(permission)

            if has_result_read_permission:
                action_object = action_object.private
            else:
                action_object = action_object.mock

        action_object.syft_point_to(context.server.id)

        return action_object

    @service_method(
        path="action.is_resolved", name="is_resolved", roles=GUEST_ROLE_LEVEL
    )
    def is_resolved(
        self,
        context: AuthedServiceContext,
        uid: UID,
    ) -> bool:
        """Get an object from the action store"""
        obj = self._get(context, uid).unwrap()

        if obj.is_link:
            result = self.resolve_links(
                context, obj.syft_action_data.action_object_id.id
            ).unwrap()
            return result.syft_resolved

        # If it's a leaf but not resolved yet, return false
        if not obj.syft_resolved:
            return False

        # If it's not an action data link or non resolved (empty). It's resolved
        return True

    @as_result(StashException, NotFoundException)
    def resolve_links(
        self,
        context: AuthedServiceContext,
        uid: UID,
        twin_mode: TwinMode = TwinMode.PRIVATE,
    ) -> ActionObject:
        """Get an object from the action store"""
        # If user has permission to get the object / object exists
        result = self.stash.get(uid=uid, credentials=context.credentials).unwrap()

        # If it's not a leaf
        if result.is_link:
            return self.resolve_links(
                context, result.syft_action_data.action_object_id.id, twin_mode
            ).unwrap()

        # If it's a leaf
        return result

    @service_method(path="action.get", name="get", roles=GUEST_ROLE_LEVEL)
    def get(
        self,
        context: AuthedServiceContext,
        uid: UID,
        twin_mode: TwinMode = TwinMode.PRIVATE,
        resolve_nested: bool = True,
    ) -> ActionObject | TwinObject:
        """Get an object from the action store"""
        return self._get(
            context, uid, twin_mode, resolve_nested=resolve_nested
        ).unwrap()

    @as_result(StashException, NotFoundException, SyftException)
    def _get(
        self,
        context: AuthedServiceContext,
        uid: UID,
        twin_mode: TwinMode = TwinMode.PRIVATE,
        has_permission: bool = False,
        resolve_nested: bool = True,
    ) -> ActionObject | TwinObject:
        """Get an object from the action store"""
        obj = self.stash.get(
            uid=uid, credentials=context.credentials, has_permission=has_permission
        ).unwrap()

        # TODO: Is this necessary?
        if context.server is None:
            raise SyftException(public_message=f"Server not found. Context: {context}")
        obj._set_obj_location_(
            context.server.id,
            context.credentials,
        )

        # Resolve graph links
        if not isinstance(obj, TwinObject) and resolve_nested and obj.is_link:  # type: ignore [unreachable]
            # if not self.is_resolved(  # type: ignore [unreachable]
            #     context, obj.syft_action_data.action_object_id.id
            # ):
            #     raise SyftException(public_message="This object is not resolved yet.")

            return self.resolve_links(  # type: ignore
                context, obj.syft_action_data.action_object_id.id, twin_mode
            ).unwrap()

        if isinstance(obj, TwinObject):
            if twin_mode == TwinMode.PRIVATE:
                obj = obj.private
                obj.syft_point_to(context.server.id)
            elif twin_mode == TwinMode.MOCK:
                obj = obj.mock
                obj.syft_point_to(context.server.id)
            else:
                obj.mock.syft_point_to(context.server.id)
                obj.private.syft_point_to(context.server.id)

        return obj

    @service_method(
        path="action.get_pointer", name="get_pointer", roles=GUEST_ROLE_LEVEL
    )
    def get_pointer(
        self, context: AuthedServiceContext, uid: UID
    ) -> ActionObjectPointer:
        """Get a pointer from the action store"""
        obj = self.stash.get_pointer(
            uid=uid, credentials=context.credentials, server_uid=context.server.id
        ).unwrap()

        obj._set_obj_location_(
            context.server.id,
            context.credentials,
        )

        return obj

    @service_method(path="action.get_mock", name="get_mock", roles=GUEST_ROLE_LEVEL)
    def get_mock(self, context: AuthedServiceContext, uid: UID) -> SyftObject:
        """Get a pointer from the action store"""
        return self.stash.get_mock(credentials=context.credentials, uid=uid).unwrap()

    @service_method(
        path="action.has_storage_permission",
        name="has_storage_permission",
        roles=GUEST_ROLE_LEVEL,
    )
    def has_storage_permission(self, context: AuthedServiceContext, uid: UID) -> bool:
        return self.stash.has_storage_permission(
            StoragePermission(uid=uid, server_uid=context.server.id)
        )

    def has_read_permission(self, context: AuthedServiceContext, uid: UID) -> bool:
        return self.stash.has_permissions(
            [ActionObjectREAD(uid=uid, credentials=context.credentials)]
        )

    # not a public service endpoint
    @as_result(SyftException)
    def _user_code_execute(
        self,
        context: AuthedServiceContext,
        code_item: UserCode,
        kwargs: dict[str, Any],
        result_id: UID | None = None,
    ) -> ActionObjectPointer:
        override_execution_permission = (
            context.has_execute_permissions or context.role == ServiceRole.ADMIN
        )
        input_policy = code_item.get_input_policy(context)
        output_policy = code_item.get_output_policy(context)

        # Unwrap nested ActionObjects
        for _, arg in kwargs.items():
            self.flatten_action_arg(context, arg) if isinstance(arg, UID) else None

        if not override_execution_permission:
            if input_policy is None:
                if not code_item.is_output_policy_approved(context).unwrap():
                    raise SyftException(
                        public_message="Execution denied: Your code is waiting for approval"
                    )
                raise SyftException(
                    public_message=f"No input policy defined for user code: {code_item.id}"
                )

            # validate input policy, raises if not valid
            input_policy.is_valid(
                context=context,
                usr_input_kwargs=kwargs,
            )

            # Filter input kwargs based on policy
            filtered_kwargs = input_policy.filter_kwargs(
                kwargs=kwargs,
                context=context,
            )
        else:
            filtered_kwargs = retrieve_from_db(kwargs, context).unwrap()

        if hasattr(input_policy, "transform_kwargs"):
            filtered_kwargs = input_policy.transform_kwargs(  # type: ignore
                context,
                filtered_kwargs,
            ).unwrap()

        # update input policy to track any input state
        has_twin_inputs = False

        real_kwargs = {}
        for key, kwarg_value in filtered_kwargs.items():
            if isinstance(kwarg_value, TwinObject):
                has_twin_inputs = True
            real_kwargs[key] = kwarg_value

        result_id = UID() if result_id is None else result_id

        try:
            if not has_twin_inputs:
                # no twins
                # allow python types from inputpolicy
                filtered_kwargs = filter_twin_kwargs(
                    real_kwargs, twin_mode=TwinMode.NONE, allow_python_types=True
                ).unwrap()
                exec_result = execute_byte_code(code_item, filtered_kwargs, context)
                if exec_result.errored:
                    raise SyftException(public_message=exec_result.safe_error_message)

                if output_policy:
                    exec_result.result = output_policy.apply_to_output(
                        context,
                        exec_result.result,
                        update_policy=not override_execution_permission,
                    )
                code_item.output_policy = output_policy  # type: ignore
                context.server.services.user_code.update_code_state(context, code_item)
                if isinstance(exec_result.result, ActionObject):
                    result_action_object = ActionObject.link(
                        result_id=result_id, pointer_id=exec_result.result.id
                    )
                else:
                    result_action_object = wrap_result(result_id, exec_result.result)
            else:
                # twins
                private_kwargs = filter_twin_kwargs(
                    real_kwargs, twin_mode=TwinMode.PRIVATE, allow_python_types=True
                ).unwrap()
                private_exec_result = execute_byte_code(
                    code_item, private_kwargs, context
                )
                if private_exec_result.errored:
                    raise SyftException(
                        public_message=private_exec_result.safe_error_message
                    )

                if output_policy:
                    private_exec_result.result = output_policy.apply_to_output(
                        context,
                        private_exec_result.result,
                        update_policy=not override_execution_permission,
                    )
                code_item.output_policy = output_policy  # type: ignore
                context.server.services.user_code.update_code_state(context, code_item)
                result_action_object_private = wrap_result(
                    result_id, private_exec_result.result
                )

                mock_kwargs = filter_twin_kwargs(
                    real_kwargs, twin_mode=TwinMode.MOCK, allow_python_types=True
                ).unwrap()
                # relative
                from .action_data_empty import ActionDataEmpty

                if any(isinstance(v, ActionDataEmpty) for v in mock_kwargs.values()):
                    mock_exec_result_obj = ActionDataEmpty()
                else:
                    mock_exec_result = execute_byte_code(
                        code_item, mock_kwargs, context
                    )

                    if mock_exec_result.errored:
                        raise SyftException(
                            public_message=mock_exec_result.safe_error_message
                        )

                    if output_policy:
                        mock_exec_result.result = output_policy.apply_to_output(
                            context, mock_exec_result.result, update_policy=False
                        )
                    mock_exec_result_obj = mock_exec_result.result

                result_action_object_mock = wrap_result(result_id, mock_exec_result_obj)

                result_action_object = TwinObject(
                    id=result_id,
                    private_obj=result_action_object_private,
                    mock_obj=result_action_object_mock,
                )
        except Exception as e:
            raise SyftException.from_exception(e)

        return result_action_object

    # def raise_for_failed_execution(self, output: ExecutionOutput):
    #     if output.errored:
    #         raise SyftException(public_message="Execution of usercode failed, ask admin",
    #                                 private_message=output.stdout + "\n" + output.stderr)

    @as_result(SyftException)
    def set_result_to_store(
        self,
        result_action_object: ActionObject | TwinObject,
        context: AuthedServiceContext,
        output_policy: OutputPolicy | None = None,
        has_result_read_permission: bool = False,
    ) -> ActionObject:
        result_id = result_action_object.id
        # result_blob_id = result_action_object.syft_blob_storage_entry_id

        if output_policy is not None:
            output_readers = (
                output_policy.output_readers
                if not context.has_execute_permissions
                else []
            )
        else:
            output_readers = []

        # If flag is True, user has read permissions to the results in BlobStore
        if has_result_read_permission:
            output_readers.append(context.credentials)

        read_permission = ActionPermission.READ

        result_action_object._set_obj_location_(
            context.server.id,
            context.credentials,
        )
        blob_store_result: SyftResponseMessage = (
            result_action_object._save_to_blob_storage().unwrap()
        )
        if isinstance(blob_store_result, SyftWarning):
            logger.debug(blob_store_result.message)

        # IMPORTANT: DO THIS ONLY AFTER ._save_to_blob_storage
        if isinstance(result_action_object, TwinObject):
            result_blob_id = result_action_object.private.syft_blob_storage_entry_id
        else:
            result_blob_id = result_action_object.syft_blob_storage_entry_id  # type: ignore[unreachable]

        # pass permission information to the action store as extra kwargs
        # context.extra_kwargs = {"has_result_read_permission": True}

        # Since this just meta data about the result, they always have access to it.
        set_result = self._set(
            context,
            result_action_object,
            has_result_read_permission=True,
        ).unwrap()

        def store_permission(
            x: SyftVerifyKey | None = None,
        ) -> ActionObjectPermission:
            return ActionObjectPermission(result_id, read_permission, x)

        def blob_permission(
            x: SyftVerifyKey | None = None,
        ) -> ActionObjectPermission:
            return ActionObjectPermission(result_blob_id, read_permission, x)

        if len(output_readers) > 0:
            store_permissions = [store_permission(x) for x in output_readers]
            self.stash.add_permissions(store_permissions)

            if result_blob_id is not None:
                blob_permissions = [blob_permission(x) for x in output_readers]
                context.server.services.blob_storage.stash.add_permissions(
                    blob_permissions
                )

        return set_result

    @as_result(SyftException)
    def execute_plan(
        self,
        plan: Any,
        context: AuthedServiceContext,
        plan_kwargs: dict[str, ActionObject],
    ) -> ActionObject:
        id2inpkey = {v.id: k for k, v in plan.inputs.items()}

        for plan_action in plan.actions:
            if (
                hasattr(plan_action.remote_self, "id")
                and plan_action.remote_self.id in id2inpkey
            ):
                plan_action.remote_self = plan_kwargs[
                    id2inpkey[plan_action.remote_self.id]
                ]
            for i, arg in enumerate(plan_action.args):
                if arg in id2inpkey:
                    plan_action.args[i] = plan_kwargs[id2inpkey[arg]]

            for k, arg in enumerate(plan_action.kwargs):
                if arg in id2inpkey:
                    plan_action.kwargs[k] = plan_kwargs[id2inpkey[arg]]

        for plan_action in plan.actions:
            self.execute(context, plan_action)
        result_id = plan.outputs[0].id
        return self._get(
            context, result_id, TwinMode.NONE, has_permission=True
        ).unwrap()

    @as_result(SyftException)
    def call_function(
        self, context: AuthedServiceContext, action: Action
    ) -> ActionObject:
        # run function/class init
        _user_lib_config_registry = UserLibConfigRegistry.from_user(context.credentials)
        absolute_path = f"{action.path}.{action.op}"
        if absolute_path in _user_lib_config_registry:
            # TODO: implement properly
            # Now we are assuming its a function/class
            return execute_callable(self, context, action).unwrap()
        else:
            raise SyftException(
                public_message=f"Failed executing {action}. You have no permission for {absolute_path}"
            )

    @as_result(SyftException)
    def set_attribute(
        self,
        context: AuthedServiceContext,
        action: Action,
        resolved_self: ActionObject | TwinObject,
    ) -> TwinObject:
        args, _ = resolve_action_args(action, context, self).unwrap(
            public_message=f"Failed executing action {action} (could not resolve args)"
        )
        if not isinstance(args[0], ActionObject):
            raise SyftException(
                public_message=(
                    f"Failed executing action {action} setattribute requires"
                    " a non-twin string as first argument"
                )
            )
        name = args[0].syft_action_data
        # dont do the whole filtering dance with the name
        args = [args[1]]

        if isinstance(resolved_self, TwinObject):
            # todo, create copy?
            private_args = filter_twin_args(args, twin_mode=TwinMode.PRIVATE).unwrap()
            private_val = private_args[0]
            setattr(resolved_self.private.syft_action_data, name, private_val)
            # todo: what do we use as data for the mock here?
            # depending on permisisons?
            public_args = filter_twin_args(args, twin_mode=TwinMode.MOCK).unwrap()
            public_val = public_args[0]
            setattr(resolved_self.mock.syft_action_data, name, public_val)
            return TwinObject(
                id=action.result_id,
                private_obj=ActionObject.from_obj(
                    resolved_self.private.syft_action_data
                ),
                private_obj_id=action.result_id,
                mock_obj=ActionObject.from_obj(resolved_self.mock.syft_action_data),
                mock_obj_id=action.result_id,
            )

        else:
            # TODO: Implement for twinobject args
            args = filter_twin_args(args, twin_mode=TwinMode.NONE).unwrap()  # type: ignore[unreachable]
            val = args[0]
            setattr(resolved_self.syft_action_data, name, val)
            return (ActionObject.from_obj(resolved_self.syft_action_data),)
            # todo: permissions
            # setattr(resolved_self.syft_action_data, name, val)
            # val = resolved_self.syft_action_data
            # result_action_object = Ok(wrap_result(action.result_id, val))

    @as_result(SyftException)
    def get_attribute(
        self, action: Action, resolved_self: ActionObject | TwinObject
    ) -> TwinObject | ActionObject:
        if isinstance(resolved_self, TwinObject):
            private_result = getattr(resolved_self.private.syft_action_data, action.op)
            mock_result = getattr(resolved_self.mock.syft_action_data, action.op)
            return TwinObject(
                id=action.result_id,
                private_obj=ActionObject.from_obj(private_result),
                private_obj_id=action.result_id,
                mock_obj=ActionObject.from_obj(mock_result),
                mock_obj_id=action.result_id,
            )
        else:
            val = getattr(resolved_self.syft_action_data, action.op)  # type: ignore[unreachable]
            return wrap_result(action.result_id, val)

    @as_result(SyftException)
    def call_method(
        self,
        context: AuthedServiceContext,
        action: Action,
        resolved_self: ActionObject | TwinObject,
    ) -> TwinObject | Any:
        if isinstance(resolved_self, TwinObject):
            # method
            private_result = execute_object(
                self,
                context,
                resolved_self.private,
                action,
                twin_mode=TwinMode.PRIVATE,
            ).unwrap(public_message=f"Failed executing action {action}")
            mock_result = execute_object(
                self, context, resolved_self.mock, action, twin_mode=TwinMode.MOCK
            ).unwrap(public_message=f"Failed executing action {action}")

            return TwinObject(
                id=action.result_id,
                private_obj=private_result,
                private_obj_id=action.result_id,
                mock_obj=mock_result,
                mock_obj_id=action.result_id,
            )
        else:
            return execute_object(self, context, resolved_self, action).unwrap()  # type:ignore[unreachable]

    as_result(SyftException)

    def unwrap_nested_actionobjects(
        self, context: AuthedServiceContext, data: Any
    ) -> Any:
        """recursively unwraps nested action objects"""

        if isinstance(data, list):
            return [self.unwrap_nested_actionobjects(context, obj) for obj in data]

        if isinstance(data, dict):
            return {
                key: self.unwrap_nested_actionobjects(context, obj)
                for key, obj in data.items()
            }

        if isinstance(data, ActionObject):
            res = self.get(context=context, uid=data.id)

            nested_res = res.syft_action_data

            if isinstance(nested_res, ActionObject):
                raise SyftException(
                    public_message="More than double nesting of ActionObjects is currently not supported"
                )

            return nested_res

        return data

    def contains_nested_actionobjects(self, data: Any) -> bool:
        """
        returns if this is a list/set/dict that contains ActionObjects
        """

        def unwrap_collection(col: set | dict | list) -> [Any]:  # type: ignore
            return_values = []
            if isinstance(col, dict):
                values = list(col.values()) + list(col.keys())
            else:
                values = list(col)
            for v in values:
                if isinstance(v, list | dict | set):
                    return_values += unwrap_collection(v)
                else:
                    return_values.append(v)
            return return_values

        if isinstance(data, list | dict | set):
            values = unwrap_collection(data)
            has_action_object = any(isinstance(x, ActionObject) for x in values)
            return has_action_object
        elif isinstance(data, ActionObject):
            return True
        return False

    def flatten_action_arg(self, context: AuthedServiceContext, arg: UID) -> None:
        """ "If the argument is a collection (of collections) of ActionObjects,
        We want to flatten the collection and upload a new ActionObject that contins
        its values. E.g. [[ActionObject1, ActionObject2],[ActionObject3, ActionObject4]]
        -> [[value1, value2],[value3, value4]]
        """
        root_context = context.as_root_context()

        action_object = self.get(context=root_context, uid=arg)
        data = action_object.syft_action_data

        if self.contains_nested_actionobjects(data):
            new_data = self.unwrap_nested_actionobjects(context, data)
            # Update existing action object with the new flattened data
            action_object.syft_action_data_cache = new_data

            # we should create this with the permissions as the old object
            # currently its using the client verify key on the object
            action_object._save_to_blob_storage().unwrap()
            # we should create this with the permissions of the old object
            self._set(
                context=root_context,
                action_object=action_object,
            ).unwrap()

        return None

    @service_method(path="action.execute", name="execute", roles=GUEST_ROLE_LEVEL)
    def execute(self, context: AuthedServiceContext, action: Action) -> ActionObject:
        """Execute an operation on objects in the action store"""
        # relative
        from .plan import Plan

        if action.action_type == ActionType.CREATEOBJECT:
            result_action_object = action.create_object
        elif action.action_type == ActionType.SYFTFUNCTION:
            kwarg_ids = {}
            for k, v in action.kwargs.items():
                # transform lineage ids into ids
                kwarg_ids[k] = v.id
            return context.server.services.user_code._call(  # type: ignore[union-attr]
                context, action.user_code_id, action.result_id, **kwarg_ids
            ).unwrap()
        elif action.action_type == ActionType.FUNCTION:
            result_action_object = self.call_function(context, action).unwrap()
        else:
            resolved_self = self._get(
                context=context,
                uid=action.remote_self,
                twin_mode=TwinMode.NONE,
                has_permission=True,
            ).unwrap(
                public_message=f"Failed executing action {action}, could not resolve self: {action.remote_self}"
            )
            if action.op == "__call__" and resolved_self.syft_action_data_type == Plan:
                result_action_object = self.execute_plan(
                    plan=resolved_self.syft_action_data,
                    context=context,
                    plan_kwargs=action.kwargs,
                ).unwrap()
            elif action.action_type == ActionType.SETATTRIBUTE:
                result_action_object = self.set_attribute(
                    context, action, resolved_self
                ).unwrap()
            elif action.action_type == ActionType.GETATTRIBUTE:
                result_action_object = self.get_attribute(
                    action, resolved_self
                ).unwrap()
            elif action.action_type == ActionType.METHOD:
                result_action_object = self.call_method(
                    context, action, resolved_self
                ).unwrap()
            else:
                raise SyftException(public_message="unknown action")

        # check if we have read permissions on the result
        has_result_read_permission = self.has_read_permission_for_action_result(
            context, action
        )
        result_action_object._set_obj_location_(  # type: ignore[union-attr]
            context.server.id,
            context.credentials,
        )
        blob_store_result = result_action_object._save_to_blob_storage().unwrap()  # type: ignore[union-attr]
        # pass permission information to the action store as extra kwargs
        context.extra_kwargs = {
            "has_result_read_permission": has_result_read_permission
        }
        if isinstance(blob_store_result, SyftWarning):
            logger.debug(blob_store_result.message)
        set_result = self._set(
            context,
            result_action_object,
        )
        set_result = set_result.unwrap(
            public_message=f"Failed executing action {action}"
        )

        return set_result

    def has_read_permission_for_action_result(
        self, context: AuthedServiceContext, action: Action
    ) -> bool:
        action_obj_ids = (
            action.args + list(action.kwargs.values()) + [action.remote_self]
        )
        permissions = [
            ActionObjectREAD(uid=_id, credentials=context.credentials)
            for _id in action_obj_ids
        ]
        return self.stash.has_permissions(permissions)

    @service_method(path="action.exists", name="exists", roles=GUEST_ROLE_LEVEL)
    def exists(self, context: AuthedServiceContext, obj_id: UID) -> bool:
        """Checks if the given object id exists in the Action Store"""
        return self.stash.exists(context.credentials, obj_id)

    @service_method(
        path="action.delete",
        name="delete",
        roles=ADMIN_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def delete(
        self, context: AuthedServiceContext, uid: UID, soft_delete: bool = False
    ) -> SyftSuccess:
        obj = self.stash.get(uid=uid, credentials=context.credentials).unwrap()

        return_msg = []

        # delete any associated blob storage entry object to the action object
        blob_del_res = self._delete_blob_storage_entry(
            context=context, obj=obj
        ).unwrap()
        return_msg.append(blob_del_res.message)

        # delete the action object from the action store
        store_del_res = self._delete_from_action_store(
            context=context, uid=obj.id, soft_delete=soft_delete
        ).unwrap()

        return_msg.append(store_del_res.message)
        return SyftSuccess(message="\n".join(return_msg))

    @as_result(SyftException)
    def _delete_blob_storage_entry(
        self,
        context: AuthedServiceContext,
        obj: TwinObject | ActionObject,
    ) -> SyftSuccess:
        deleted_blob_ids = []

        if isinstance(obj, ActionObject) and obj.syft_blob_storage_entry_id:
            context.server.services.blob_storage.delete(
                context=context, uid=obj.syft_blob_storage_entry_id
            )
            deleted_blob_ids.append(obj.syft_blob_storage_entry_id)

        if isinstance(obj, TwinObject):
            if obj.private.syft_blob_storage_entry_id:
                context.server.services.blob_storage.delete(
                    context=context, uid=obj.private.syft_blob_storage_entry_id
                )
                deleted_blob_ids.append(obj.private.syft_blob_storage_entry_id)

            if obj.mock.syft_blob_storage_entry_id:
                context.server.services.blob_storage.delete(
                    context=context, uid=obj.mock.syft_blob_storage_entry_id
                )
                deleted_blob_ids.append(obj.mock.syft_blob_storage_entry_id)

        message = f"Deleted blob storage entries: {', '.join(str(blob_id) for blob_id in deleted_blob_ids)}"
        return SyftSuccess(message=message)

    @as_result(SyftException)
    def _delete_from_action_store(
        self,
        context: AuthedServiceContext,
        uid: UID,
        soft_delete: bool = False,
    ) -> SyftSuccess:
        if soft_delete:
            obj = self.stash.get(uid=uid, credentials=context.credentials).unwrap()

            if isinstance(obj, TwinObject):
                self._soft_delete_action_obj(
                    context=context, action_obj=obj.private
                ).unwrap()
                self._soft_delete_action_obj(
                    context=context, action_obj=obj.mock
                ).unwrap()
            if isinstance(obj, ActionObject):
                self._soft_delete_action_obj(context=context, action_obj=obj).unwrap()
        else:
            self.stash.delete_by_uid(credentials=context.credentials, uid=uid).unwrap()

        return SyftSuccess(message=f"Action object with uid '{uid}' deleted.")

    @as_result(SyftException)
    def _soft_delete_action_obj(
        self, context: AuthedServiceContext, action_obj: ActionObject
    ) -> ActionObject:
        action_obj.syft_action_data_cache = None
        action_obj._save_to_blob_storage().unwrap()
        return self._set(
            context=context,
            action_object=action_obj,
        ).unwrap()


@as_result(SyftException)
def resolve_action_args(
    action: Action, context: AuthedServiceContext, service: ActionService
) -> tuple[list, bool]:
    has_twin_inputs = False
    args = []
    for arg_id in action.args:
        arg_value = service._get(
            context=context, uid=arg_id, twin_mode=TwinMode.NONE, has_permission=True
        ).unwrap()
        if isinstance(arg_value, TwinObject):
            has_twin_inputs = True
        args.append(arg_value)
    return args, has_twin_inputs


@as_result(SyftException)
def resolve_action_kwargs(
    action: Action, context: AuthedServiceContext, service: ActionService
) -> tuple[dict, bool]:
    has_twin_inputs = False
    kwargs = {}
    for key, arg_id in action.kwargs.items():
        kwarg_value = service._get(
            context=context, uid=arg_id, twin_mode=TwinMode.NONE, has_permission=True
        ).unwrap()
        if isinstance(kwarg_value, TwinObject):
            has_twin_inputs = True
        kwargs[key] = kwarg_value
    return kwargs, has_twin_inputs


@as_result(SyftException)
def execute_callable(
    service: ActionService,
    context: AuthedServiceContext,
    action: Action,
) -> ActionObject:
    args, has_arg_twins = resolve_action_args(action, context, service).unwrap()
    kwargs, has_kwargs_twins = resolve_action_kwargs(action, context, service).unwrap()
    has_twin_inputs = has_arg_twins or has_kwargs_twins
    # 🔵 TODO 10: Get proper code From old RunClassMethodAction to ensure the function
    # is not bound to the original object or mutated

    # TODO: get from CMPTree is probably safer
    def _get_target_callable(path: str, op: str) -> Any:
        path_elements = path.split(".")
        res = importlib.import_module(path_elements[0])
        for p in path_elements[1:]:
            res = getattr(res, p)
        res = getattr(res, op)
        return res

    target_callable = _get_target_callable(action.path, action.op)

    result = None
    if not target_callable:
        raise SyftException(public_message="No target callable found")

    if not has_twin_inputs:
        # if twin_mode == TwinMode.NONE and not has_twin_inputs:
        twin_mode = TwinMode.NONE
        # no twins
        filtered_args = filter_twin_args(args, twin_mode=twin_mode).unwrap()
        filtered_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode).unwrap()
        result = target_callable(*filtered_args, **filtered_kwargs)
        result_action_object = wrap_result(action.result_id, result)
    else:
        twin_mode = TwinMode.PRIVATE
        private_args = filter_twin_args(args, twin_mode=twin_mode).unwrap()
        private_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode).unwrap()
        private_result = target_callable(*private_args, **private_kwargs)
        result_action_object_private = wrap_result(action.result_id, private_result)

        twin_mode = TwinMode.MOCK
        mock_args = filter_twin_args(args, twin_mode=twin_mode).unwrap()
        mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode).unwrap()
        mock_result = target_callable(*mock_args, **mock_kwargs)
        result_action_object_mock = wrap_result(action.result_id, mock_result)

        result_action_object = TwinObject(
            id=action.result_id,
            private_obj=result_action_object_private,
            mock_obj=result_action_object_mock,
        )

    return result_action_object


@as_result(SyftException)
def execute_object(
    service: ActionService,
    context: AuthedServiceContext,
    resolved_self: ActionObject,
    action: Action,
    twin_mode: TwinMode = TwinMode.NONE,
) -> TwinObject | ActionObject:
    unboxed_resolved_self = resolved_self.syft_action_data
    args, has_arg_twins = resolve_action_args(action, context, service).unwrap()

    kwargs, has_kwargs_twins = resolve_action_kwargs(action, context, service).unwrap()
    has_twin_inputs = has_arg_twins or has_kwargs_twins

    # 🔵 TODO 10: Get proper code From old RunClassMethodAction to ensure the function
    # is not bound to the original object or mutated
    target_method = getattr(unboxed_resolved_self, action.op, None)
    result = None

    if not target_method:
        raise SyftException(public_message="could not find target method")
    if twin_mode == TwinMode.NONE and not has_twin_inputs:
        # no twins
        filtered_args = filter_twin_args(args, twin_mode=twin_mode).unwrap()
        filtered_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode).unwrap()
        result = target_method(*filtered_args, **filtered_kwargs)
        result_action_object = wrap_result(action.result_id, result)
    elif twin_mode == TwinMode.NONE and has_twin_inputs:
        # self isn't a twin but one of the inputs is
        private_args = filter_twin_args(args, twin_mode=TwinMode.PRIVATE).unwrap()
        private_kwargs = filter_twin_kwargs(kwargs, twin_mode=TwinMode.PRIVATE).unwrap()
        private_result = target_method(*private_args, **private_kwargs)
        result_action_object_private = wrap_result(action.result_id, private_result)

        mock_args = filter_twin_args(args, twin_mode=TwinMode.MOCK).unwrap()
        mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=TwinMode.MOCK).unwrap()
        mock_result = target_method(*mock_args, **mock_kwargs)
        result_action_object_mock = wrap_result(action.result_id, mock_result)

        result_action_object = TwinObject(
            id=action.result_id,
            private_obj=result_action_object_private,
            mock_obj=result_action_object_mock,
        )
    elif twin_mode == twin_mode.PRIVATE:  # type:ignore
        # twin private path
        private_args = filter_twin_args(args, twin_mode=twin_mode).unwrap()  # type:ignore[unreachable]
        private_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode).unwrap()
        result = target_method(*private_args, **private_kwargs)
        result_action_object = wrap_result(action.result_id, result)
    elif twin_mode == twin_mode.MOCK:  # type:ignore
        # twin mock path
        mock_args = filter_twin_args(args, twin_mode=twin_mode).unwrap()  # type:ignore[unreachable]
        mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode).unwrap()
        target_method = getattr(unboxed_resolved_self, action.op, None)
        result = target_method(*mock_args, **mock_kwargs)
        result_action_object = wrap_result(action.result_id, result)
    else:
        raise SyftException(
            public_message=f"Bad combination of: twin_mode: {twin_mode} and has_twin_inputs: {has_twin_inputs}"
        )

    return result_action_object


def wrap_result(result_id: UID, result: Any) -> ActionObject:
    # 🟡 TODO 11: Figure out how we want to store action object results
    action_type = action_type_for_type(result)
    result_action_object = action_type(id=result_id, syft_action_data_cache=result)
    return result_action_object


@as_result(SyftException)
def filter_twin_args(args: list[Any], twin_mode: TwinMode) -> Any:
    filtered = []
    for arg in args:
        if isinstance(arg, TwinObject):
            if twin_mode == TwinMode.PRIVATE:
                filtered.append(arg.private.syft_action_data)
            elif twin_mode == TwinMode.MOCK:
                filtered.append(arg.mock.syft_action_data)
            else:
                raise SyftException(
                    public_message=f"Filter can only use {TwinMode.PRIVATE} or {TwinMode.MOCK}"
                )
        else:
            filtered.append(arg.syft_action_data)
    return filtered


@as_result(SyftException)
def filter_twin_kwargs(
    kwargs: dict, twin_mode: TwinMode, allow_python_types: bool = False
) -> Any:
    filtered = {}
    for k, v in kwargs.items():
        if isinstance(v, TwinObject):
            if twin_mode == TwinMode.PRIVATE:
                filtered[k] = v.private.syft_action_data
            elif twin_mode == TwinMode.MOCK:
                filtered[k] = v.mock.syft_action_data
            else:
                raise SyftException(
                    public_message=f"Filter can only use {TwinMode.PRIVATE} or {TwinMode.MOCK}"
                )
        else:
            if isinstance(v, ActionObject):
                filtered[k] = v.syft_action_data
            elif (
                isinstance(v, str | int | float | dict | CustomEndpointActionObject)
                and allow_python_types
            ):
                filtered[k] = v
            else:
                raise SyftException(
                    public_message=f"unexepected value {v} passed to filtered twin kwargs"
                )
    return filtered


TYPE_TO_SERVICE[ActionObject] = ActionService
TYPE_TO_SERVICE[TwinObject] = ActionService
TYPE_TO_SERVICE[AnyActionObject] = ActionService

SERVICE_TO_TYPES[ActionService].update({ActionObject, TwinObject, AnyActionObject})



# .api.services.notifications.settings() is how the server itself would dispatch notifications.
# .api.services.notifications.user_settings() sets if a specific user wants or not to receive notifications.
# Class NotifierSettings holds both pieces of info.
# Users will get notification x where x in {email, slack, sms, app} if three things are set to True:
# 1) .....settings().active
# 2) .....settings().x_enabled
# 2) .....user_settings().x

# stdlib
from collections.abc import Callable
from datetime import datetime
import logging
from typing import Any
from typing import TypeVar

# third party
from pydantic import BaseModel

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.syft_migration import migrate
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SYFT_OBJECT_VERSION_2
from ...types.syft_object import SYFT_OBJECT_VERSION_3
from ...types.syft_object import SyftObject
from ...types.transforms import drop
from ...types.transforms import make_set_default
from ..context import AuthedServiceContext
from ..notification.notifications import Notification
from ..response import SyftError
from ..response import SyftSuccess
from .notifier_enums import NOTIFICATION_FREQUENCY
from .notifier_enums import NOTIFIERS
from .smtp_client import SMTPClient

logger = logging.getLogger(__name__)


class BaseNotifier(BaseModel):
    @as_result(SyftException)
    def send(
        self, context: AuthedServiceContext, notification: Notification
    ) -> SyftSuccess:
        raise SyftException(public_message="Not implemented")

    @as_result(SyftException)
    def send_batches(
        self, context: AuthedServiceContext, notification_queue: list[Notification]
    ) -> SyftSuccess:
        raise SyftException(public_message="Not implemented")


TBaseNotifier = TypeVar("TBaseNotifier", bound=BaseNotifier)


@serializable()
class UserNotificationActivity(SyftObject):
    __canonical_name__ = "UserNotificationActivity"
    __version__ = SYFT_OBJECT_VERSION_1
    count: int = 1
    date: datetime = datetime.now()


@serializable(canonical_name="EmailNotifier", version=1)
class EmailNotifier(BaseNotifier):
    smtp_client: SMTPClient | None = None
    sender: str = ""

    def __init__(
        self,
        **data: Any,
    ) -> None:
        super().__init__(**data)
        self.sender = data.get("sender", "")
        self.smtp_client = SMTPClient(
            server=data.get("server", ""),
            port=int(data.get("port", 587)),
            username=data.get("username", ""),
            password=data.get("password", ""),
        )

    @classmethod
    def check_credentials(
        cls,
        username: str,
        password: str,
        server: str,
        port: int = 587,
    ) -> bool:
        try:
            SMTPClient.check_credentials(
                server=server,
                port=port,
                username=username,
                password=password,
            )
            return True
        except Exception:
            logger.exception("Credentials validation failed")
            return False

    @as_result(SyftException)
    def send_batches(
        self, context: AuthedServiceContext, notification_queue: list[Notification]
    ) -> SyftSuccess | SyftError:
        subject = None
        receiver_email = None
        sender = None

        notification_sample = notification_queue[0]
        try:
            sender = self.sender
            receiver = context.server.services.user.get_by_verify_key(
                notification_sample.to_user_verify_key
            ).unwrap()
            if not receiver.notifications_enabled[NOTIFIERS.EMAIL]:
                return SyftSuccess(
                    message="Email notifications are disabled for this user."
                )  # TODO: Should we return an error here?
            receiver_email = receiver.email
            if notification_sample.email_template:
                subject = notification_sample.email_template.batched_email_title(
                    notifications=notification_queue, context=context
                )
                body = notification_sample.email_template.batched_email_body(
                    notifications=notification_queue, context=context
                )
            else:
                subject = notification_sample.subject
                body = notification_sample._repr_html_()

            if isinstance(receiver_email, str):
                receiver_email = [receiver_email]

            self.smtp_client.send(  # type: ignore
                sender=sender, receiver=receiver_email, subject=subject, body=body
            )
            message = f"> Sent email: {subject} to {receiver_email}"
            logging.info(message)
            return SyftSuccess(message="Email sent successfully!")
        except Exception as e:
            message = f"> Error sending email: {subject} to {receiver_email} from: {sender}. {e}"
            logger.error(message)
            return SyftError(message="Failed to send an email.")

    @as_result(SyftException)
    def send(
        self, context: AuthedServiceContext, notification: Notification
    ) -> SyftSuccess | SyftError:
        subject = None
        receiver_email = None
        sender = None
        try:
            sender = self.sender
            receiver = context.server.services.user.get_by_verify_key(
                notification.to_user_verify_key
            ).unwrap()
            if not receiver.notifications_enabled[NOTIFIERS.EMAIL]:
                return SyftSuccess(
                    message="Email notifications are disabled for this user."
                )  # TODO: Should we return an error here?
            receiver_email = receiver.email

            if notification.email_template:
                subject = notification.email_template.email_title(
                    notification, context=context
                )
                body = notification.email_template.email_body(
                    notification, context=context
                )
            else:
                subject = notification.subject
                body = notification._repr_html_()

            if isinstance(receiver_email, str):
                receiver_email = [receiver_email]

            self.smtp_client.send(  # type: ignore
                sender=sender, receiver=receiver_email, subject=subject, body=body
            )
            message = f"> Sent email: {subject} to {receiver_email}"
            print(message)
            logging.info(message)
            return SyftSuccess(message="Email sent successfully!")
        except Exception as e:
            message = f"> Error sending email: {subject} to {receiver_email} from: {sender}. {e}"
            logger.error(message)
            return SyftError(message="Failed to send an email.")
            # raise SyftException.from_exception(
            #     exc,
            #     public_message=(
            #         "Some notifications failed to be delivered."
            #         " Please check the health of the mailing server."
            #     ),
            # )


@serializable()
class NotificationPreferences(SyftObject):
    __canonical_name__ = "NotificationPreferences"
    __version__ = SYFT_OBJECT_VERSION_1
    __repr_attrs__ = [
        "email",
        "sms",
        "slack",
        "app",
    ]

    email: bool = False
    sms: bool = False
    slack: bool = False
    app: bool = False


@serializable()
class NotifierSettingsV1(SyftObject):
    __canonical_name__ = "NotifierSettings"
    __version__ = SYFT_OBJECT_VERSION_1
    __repr_attrs__ = [
        "active",
        "email_enabled",
    ]
    active: bool = False

    notifiers: dict[NOTIFIERS, type[TBaseNotifier]] = {
        NOTIFIERS.EMAIL: EmailNotifier,
    }

    notifiers_status: dict[NOTIFIERS, bool] = {
        NOTIFIERS.EMAIL: True,
        NOTIFIERS.SMS: False,
        NOTIFIERS.SLACK: False,
        NOTIFIERS.APP: False,
    }

    email_sender: str | None = ""
    email_server: str | None = ""
    email_port: int | None = 587
    email_username: str | None = ""
    email_password: str | None = ""


@serializable()
class NotifierSettingsV2(SyftObject):
    __canonical_name__ = "NotifierSettings"
    __version__ = SYFT_OBJECT_VERSION_2
    __repr_attrs__ = [
        "active",
        "email_enabled",
    ]
    active: bool = False
    # Flag to identify which notification is enabled
    # For now, consider only the email notification
    # In future, Admin, must be able to have a better
    # control on diff notifications.

    notifiers: dict[NOTIFIERS, type[TBaseNotifier]] = {
        NOTIFIERS.EMAIL: EmailNotifier,
    }

    notifiers_status: dict[NOTIFIERS, bool] = {
        NOTIFIERS.EMAIL: True,
        NOTIFIERS.SMS: False,
        NOTIFIERS.SLACK: False,
        NOTIFIERS.APP: False,
    }

    email_sender: str | None = ""
    email_server: str | None = ""
    email_port: int | None = 587
    email_username: str | None = ""
    email_password: str | None = ""

    email_activity: dict[str, dict[SyftVerifyKey, UserNotificationActivity]] = {}
    email_rate_limit: dict[str, int] = {}


@serializable()
class EmailFrequency(SyftObject):
    __canonical_name__ = "EmailFrequency"
    __version__ = SYFT_OBJECT_VERSION_1

    frequency: NOTIFICATION_FREQUENCY
    start_time: datetime = datetime.now()


@serializable()
class NotifierSettings(SyftObject):
    __canonical_name__ = "NotifierSettings"
    __version__ = SYFT_OBJECT_VERSION_3
    __repr_attrs__ = [
        "active",
        "email_enabled",
    ]
    active: bool = False
    # Flag to identify which notification is enabled
    # For now, consider only the email notification
    # In future, Admin, must be able to have a better
    # control on diff notifications.

    notifiers: dict[NOTIFIERS, type[TBaseNotifier]] = {
        NOTIFIERS.EMAIL: EmailNotifier,
    }

    notifiers_status: dict[NOTIFIERS, bool] = {
        NOTIFIERS.EMAIL: True,
        NOTIFIERS.SMS: False,
        NOTIFIERS.SLACK: False,
        NOTIFIERS.APP: False,
    }

    email_sender: str | None = ""
    email_server: str | None = ""
    email_port: int | None = 587
    email_username: str | None = ""
    email_password: str | None = ""
    email_frequency: dict[str, EmailFrequency] = {}
    email_queue: dict[str, dict[SyftVerifyKey, list[Notification]]] = {}
    email_activity: dict[str, dict[SyftVerifyKey, UserNotificationActivity]] = {}
    email_rate_limit: dict[str, int] = {}

    @property
    def email_enabled(self) -> bool:
        return self.notifiers_status[NOTIFIERS.EMAIL]

    @property
    def sms_enabled(self) -> bool:
        return self.notifiers_status[NOTIFIERS.SMS]

    @property
    def slack_enabled(self) -> bool:
        return self.notifiers_status[NOTIFIERS.SLACK]

    @property
    def app_enabled(self) -> bool:
        return self.notifiers_status[NOTIFIERS.APP]

    def validate_email_credentials(
        self,
        username: str,
        password: str,
        server: str,
        port: int,
    ) -> bool:
        return self.notifiers[NOTIFIERS.EMAIL].check_credentials(
            server=server,
            port=port,
            username=username,
            password=password,
        )

    @as_result(SyftException)
    def send_batched_notification(
        self,
        context: AuthedServiceContext,
        notification_queue: list[Notification],
    ) -> None:
        if len(notification_queue) == 0:
            return None
        notifier_objs: list[BaseNotifier] = self.select_notifiers(notification_queue[0])
        for notifier in notifier_objs:
            notifier.send_batches(
                context=context, notification_queue=notification_queue
            ).unwrap()
        return None

    @as_result(SyftException)
    def send_notifications(
        self,
        context: AuthedServiceContext,
        notification: Notification,
    ) -> int:
        notifier_objs: list[BaseNotifier] = self.select_notifiers(notification)

        for notifier in notifier_objs:
            notifier.send(context=context, notification=notification).unwrap()

        return len(notifier_objs)

    def select_notifiers(self, notification: Notification) -> list[BaseNotifier]:
        """
        Return a list of the notifiers enabled for the given notification"

        Args:
            notification (Notification): The notification object
        Returns:
            List[BaseNotifier]: A list of enabled notifier objects
        """
        notifier_objs = []
        for notifier_type in notification.notifier_types:
            # Check if the notifier is enabled and if it is, create the notifier object
            if (
                self.notifiers_status[notifier_type]
                and self.notifiers[notifier_type] is not None
            ):
                # If notifier is email, we need to pass the parameters
                if notifier_type == NOTIFIERS.EMAIL:
                    notifier_objs.append(
                        self.notifiers[notifier_type](  # type: ignore[misc]
                            username=self.email_username,
                            password=self.email_password,
                            sender=self.email_sender,
                            server=self.email_server,
                            port=self.email_port,
                        )
                    )
                # If notifier is not email, we just create the notifier object
                # TODO: Add the other notifiers, and its auth methods
                else:
                    notifier_objs.append(self.notifiers[notifier_type]())  # type: ignore[misc]

        return notifier_objs


@migrate(NotifierSettingsV1, NotifierSettingsV2)
def migrate_server_settings_v1_to_v2() -> list[Callable]:
    return [
        make_set_default("email_activity", {}),
        make_set_default("email_rate_limit", {}),
    ]


@migrate(NotifierSettingsV2, NotifierSettingsV1)
def migrate_server_settings_v2_to_v1() -> list[Callable]:
    # Use drop function on "notifications_enabled" attrubute
    return [drop(["email_activity"]), drop(["email_rate_limit"])]


@migrate(NotifierSettingsV2, NotifierSettings)
def migrate_server_settings_v2_to_current() -> list[Callable]:
    return [
        make_set_default("email_frequency", {}),
        make_set_default("email_queue", {}),
    ]


@migrate(NotifierSettings, NotifierSettingsV2)
def migrate_server_settings_current_to_v2() -> list[Callable]:
    # Use drop function on "notifications_enabled" attrubute
    return [drop(["email_frequency"]), drop(["email_queue"])]
